"""Test features related to both the Line and Plane."""
import pytest

from skspatial.objects import Line
from skspatial.objects import Plane


@pytest.mark.parametrize("class_spatial", [Line, Plane])
@pytest.mark.parametrize(
    ("point", "vector"),
    [([0, 0], [0, 0]), ([1, 1], [0, 0]), ([1, 1, 1], [0, 0, 0]), ([4, 5, 2, 3], [0, 0, 0, 0])],
)
def test_zero_vector_failure(class_spatial, point, vector):

    with pytest.raises(ValueError, match="The vector must not be the zero vector."):
        class_spatial(point, vector)


@pytest.mark.parametrize("class_spatial", [Line, Plane])
@pytest.mark.parametrize(("point", "vector"), [([0, 0, 1], [1, 1]), ([0, 0], [1]), ([1], [0, 1])])
def test_dimension_failure(class_spatial, point, vector):

    with pytest.raises(ValueError, match="The point and vector must have the same dimension."):
        class_spatial(point, vector)


@pytest.mark.parametrize(
    ("obj_1", "obj_2", "bool_expected"),
    [
        (Line([0, 0], [1, 0]), Line([0, 0], [1, 0]), True),
        (Line([0, 0], [1, 0]), Line([1, 0], [1, 0]), True),
        (Line([0, 0], [1, 0]), Line([-5, 0], [1, 0]), True),
        (Line([0, 0], [1, 0]), Line([-5, 0], [7, 0]), True),
        (Line([0, 0], [1, 0]), Line([-5, 0], [-20, 0]), True),
        (Line([0, 0], [1, 0]), Line([-5, 1], [1, 0]), False),
        (Plane([0, 0, 0], [0, 0, 1]), Plane([0, 0, 0], [0, 0, 1]), True),
        (Plane([0, 0, 0], [0, 0, 1]), Plane([0, 0, 0], [0, 0, 2]), True),
        (Plane([0, 0, 0], [0, 0, 1]), Plane([0, 0, 0], [0, 0, -10]), True),
        (Plane([0, 0, 0], [0, 0, 1]), Plane([0, 0, 0], [1, 0, -10]), False),
        (Line([0, 0], [1, 0]), Plane([0, 0], [1, 0]), None),
        (Plane([0, 0], [1, 0]), Line([0, 0], [1, 0]), None),
    ],
)
def test_is_close(obj_1, obj_2, bool_expected):

    if bool_expected is None:
        with pytest.raises(TypeError, match="The input must have the same type as the object."):
            obj_1.is_close(obj_2)

    else:
        assert obj_1.is_close(obj_2) == bool_expected



from math import isclose
from math import pi
from math import sqrt

import pytest

from skspatial.objects import Cylinder
from skspatial.objects import Line
from skspatial.objects import Point
from skspatial.objects import Points

LINE_DOES_NOT_INTERSECT_CYLINDER = "The line does not intersect the cylinder."
LINE_MUST_BE_3D = "The line must be 3D."


@pytest.mark.parametrize(
    ("point", "vector", "radius", "message_expected"),
    [
        ([0, 0], [1, 0, 0], 1, "The point must be 3D."),
        ([0, 0, 0], [1, 0], 1, "The vector must be 3D."),
        ([0, 0, 0], [0, 0, 0], 1, "The vector must not be the zero vector."),
        ([0, 0, 0], [0, 0, 1], 0, "The radius must be positive."),
    ],
)
def test_failure(point, vector, radius, message_expected):

    with pytest.raises(ValueError, match=message_expected):
        Cylinder(point, vector, radius)


@pytest.mark.parametrize(
    ("array_a", "array_b", "radius", "cylinder_expected"),
    [
        ([0, 0, 0], [0, 0, 1], 1, Cylinder([0, 0, 0], [0, 0, 1], 1)),
        ([0, 0, 1], [0, 0, 2], 1, Cylinder([0, 0, 1], [0, 0, 1], 1)),
        ([0, 0, 0], [1, 1, 1], 1, Cylinder([0, 0, 0], [1, 1, 1], 1)),
        ([2, 2, 2], [1, 1, 1], 5, Cylinder([2, 2, 2], [-1, -1, -1], 5)),
    ],
)
def test_from_points(array_a, array_b, radius, cylinder_expected):

    cylinder_from_points = Cylinder.from_points(array_a, array_b, radius)

    assert cylinder_from_points.vector.is_close(cylinder_expected.vector)
    assert cylinder_from_points.point.is_close(cylinder_expected.point)
    assert cylinder_from_points.radius == cylinder_expected.radius


@pytest.mark.parametrize(
    ("cylinder", "length_expected", "volume_expected"),
    [
        (Cylinder([0, 0, 0], [0, 0, 1], 1), 1, pi),
        (Cylinder([0, 0, 0], [0, 0, 1], 2), 1, 4 * pi),
        (Cylinder([0, 0, 0], [0, 0, 2], 1), 2, 2 * pi),
        (Cylinder([0, 0, 0], [0, 0, 2], 2), 2, 8 * pi),
        (Cylinder([1, 1, 1], [0, 0, 2], 2), 2, 8 * pi),
        (Cylinder([0, 0, 0], [0, 1, 1], 1), sqrt(2), sqrt(2) * pi),
        (Cylinder([0, 0, 0], [1, 1, 1], 1), sqrt(3), sqrt(3) * pi),
        (Cylinder([0, 0, 0], [5, 5, 5], 2), 5 * sqrt(3), 20 * sqrt(3) * pi),
    ],
)
def test_properties(cylinder, length_expected, volume_expected):

    assert isclose(cylinder.length(), length_expected)
    assert isclose(cylinder.volume(), volume_expected)


@pytest.mark.parametrize(
    ("cylinder", "lateral_surface_area_expected", "surface_area_expected"),
    [
        (Cylinder([0, 0, 0], [0, 0, 1], 1), 2 * pi, 4 * pi),
        (Cylinder([0, 0, 0], [0, 0, 2], 1), 4 * pi, 6 * pi),
        (Cylinder([0, 0, 0], [0, 0, 1], 2), 4 * pi, 12 * pi),
        (Cylinder([0, 0, 0], [0, 0, 2], 2), 8 * pi, 16 * pi),
        (Cylinder([0, 0, 0], [0, 0, -2], 2), 8 * pi, 16 * pi),
    ],
)
def test_surface_area(cylinder, lateral_surface_area_expected, surface_area_expected):

    assert isclose(cylinder.lateral_surface_area(), lateral_surface_area_expected)
    assert isclose(cylinder.surface_area(), surface_area_expected)


@pytest.mark.parametrize(
    ("cylinder", "point", "bool_expected"),
    [
        (Cylinder([0, 0, 0], [0, 0, 1], 1), [0, 0, 0], True),
        (Cylinder([0, 0, 0], [0, 0, 1], 1), [0, 0, 1], True),
        (Cylinder([0, 0, 0], [0, 0, 1], 1), [0, 0, 0.9], True),
        (Cylinder([0, 0, 0], [0, 0, 1], 1), [0, 0, 1.1], False),
        (Cylinder([0, 0, 0], [0, 0, 1], 1), [0, 0, -0.1], False),
        (Cylinder([0, 0, 0], [0, 0, 1], 1), [1, 0, 0], True),
        (Cylinder([0, 0, 0], [0, 0, 1], 1), [2, 0, 0], False),
        (Cylinder([0, 0, 0], [0, 0, 1], 1), [-1, 0, 0], True),
        (Cylinder([0, 0, 0], [0, 0, 1], 1), [-2, 0, 0], False),
        (Cylinder([0, 0, 0], [0, 0, 1], 1), [1, 1, 0], False),
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            [sqrt(2) / 2, sqrt(2) / 2, 0],
            True,
        ),
    ],
)
def test_cylinder_is_point_within(cylinder, point, bool_expected):

    assert cylinder.is_point_within(point) == bool_expected


@pytest.mark.parametrize(
    ("cylinder", "line", "array_expected_a", "array_expected_b"),
    [
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            Line([0, 0, 0], [1, 0, 0]),
            [-1, 0, 0],
            [1, 0, 0],
        ),
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            Line([0, 0, 0.5], [1, 0, 0]),
            [-1, 0, 0.5],
            [1, 0, 0.5],
        ),
        (
            Cylinder([0, 0, 0], [0, 0, 1], 2),
            Line([0, 0, 0], [1, 0, 0]),
            [-2, 0, 0],
            [2, 0, 0],
        ),
        (
            Cylinder([0, 0, 0], [0, 0, 5], 1),
            Line([0, 0, 0], [1, 0, 0]),
            [-1, 0, 0],
            [1, 0, 0],
        ),
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            Line([0, 0, 0], [1, 1, 0]),
            [-sqrt(2) / 2, -sqrt(2) / 2, 0],
            [sqrt(2) / 2, sqrt(2) / 2, 0],
        ),
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            Line([0, 0, 0], [1, 1, 1]),
            3 * [-sqrt(2) / 2],
            3 * [sqrt(2) / 2],
        ),
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            Line([0, -1, 0], [1, 0, 0]),
            [0, -1, 0],
            [0, -1, 0],
        ),
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            Line([0, 1, 0], [1, 0, 0]),
            [0, 1, 0],
            [0, 1, 0],
        ),
        (
            Cylinder([1, 0, 0], [0, 0, 1], 1),
            Line([0, -1, 0], [1, 0, 0]),
            [1, -1, 0],
            [1, -1, 0],
        ),
    ],
)
def test_intersect_cylinder_line(cylinder, line, array_expected_a, array_expected_b):

    point_a, point_b = cylinder.intersect_line(line, n_digits=9)

    point_expected_a = Point(array_expected_a)
    point_expected_b = Point(array_expected_b)

    assert point_a.is_close(point_expected_a)
    assert point_b.is_close(point_expected_b)


@pytest.mark.parametrize(
    ("cylinder", "line", "array_expected_a", "array_expected_b"),
    [
        # The line is parallel to the cylinder axis.
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            Line([0, 0, 0], [0, 0, 1]),
            [0, 0, 0],
            [0, 0, 1],
        ),
        # The line is perpendicular to the cylinder axis.
        (
            Cylinder([0, 0, 0], [0, 0, 2], 1),
            Line([0, 0, 1], [1, 0, 0]),
            [-1, 0, 1],
            [1, 0, 1],
        ),
        # The line touches the rim of one cylinder cap.
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            Line([1, 0, 0], [1, 0, 1]),
            [1, 0, 0],
            [1, 0, 0],
        ),
        # The line touches the edge of the lateral surface.
        (
            Cylinder([0, 0, 0], [0, 0, 2], 1),
            Line([-1, 0, 1], [0, 1, 0]),
            [-1, 0, 1],
            [-1, 0, 1],
        ),
        (
            Cylinder([0, 0, 0], [0, 0, 2], 1),
            Line([-1, 0, 1], [0, 1, 1]),
            [-1, 0, 1],
            [-1, 0, 1],
        ),
        # The line intersects one cap and the lateral surface.
        (
            Cylinder([0, 0, 0], [0, 0, 5], 1),
            Line([0, 0, 0], [1, 0, 1]),
            [0, 0, 0],
            [1, 0, 1],
        ),
        (
            Cylinder([0, 0, 0], [0, 0, 5], 1),
            Line([0, 0, 5], [1, 0, -1]),
            [0, 0, 5],
            [1, 0, 4],
        ),
    ],
)
def test_intersect_cylinder_line_with_caps(cylinder, line, array_expected_a, array_expected_b):

    point_a, point_b = cylinder.intersect_line(line, infinite=False)

    point_expected_a = Point(array_expected_a)
    point_expected_b = Point(array_expected_b)

    assert point_a.is_close(point_expected_a)
    assert point_b.is_close(point_expected_b)


@pytest.mark.parametrize(
    ("cylinder", "line", "message_expected"),
    [
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            Line([0, -2, 0], [1, 0, 0]),
            LINE_DOES_NOT_INTERSECT_CYLINDER,
        ),
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            Line([0, -2, 0], [1, 0, 1]),
            LINE_DOES_NOT_INTERSECT_CYLINDER,
        ),
        (
            Cylinder([3, 10, 4], [-1, 2, -3], 3),
            Line([0, -2, 0], [1, 0, 1]),
            LINE_DOES_NOT_INTERSECT_CYLINDER,
        ),
        (
            Cylinder([3, 10, 4], [-1, 2, -3], 3),
            Line([0, 0], [1, 0]),
            LINE_MUST_BE_3D,
        ),
        (
            Cylinder([3, 10, 4], [-1, 2, -3], 3),
            Line(4 * [0], [1, 0, 0, 0]),
            LINE_MUST_BE_3D,
        ),
    ],
)
def test_intersect_cylinder_line_failure(cylinder, line, message_expected):

    with pytest.raises(ValueError, match=message_expected):
        cylinder.intersect_line(line)


@pytest.mark.parametrize(
    ("cylinder", "line"),
    [
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            Line([0, 0, -1], [1, 0, 0]),
        ),
    ],
)
def test_intersect_cylinder_line_with_caps_failure(cylinder, line):

    message_expected = "The line does not intersect the cylinder."

    with pytest.raises(ValueError, match=message_expected):
        cylinder.intersect_line(line, infinite=False)


@pytest.mark.parametrize(
    ("cylinder", "n_along_axis", "n_angles", "points_expected"),
    [
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            1,
            1,
            [[-1, 0, 0]],
        ),
        (
            Cylinder([0, 0, 0], [0, 0, 1], 1),
            3,
            2,
            [[-1, 0, 0], [-1, 0, 0.5], [-1, 0, 1]],
        ),
        (
            Cylinder([0, 0, 0], [0.707, 0.707, 0], 1),
            1,
            3,
            [[-0.707, 0.707, 0], [0.707, -0.707, -0]],
        ),
    ],
)
def test_to_points(cylinder, n_along_axis, n_angles, points_expected):

    array_rounded = cylinder.to_points(n_along_axis=n_along_axis, n_angles=n_angles).round(3)
    points_unique = Points(array_rounded).unique()

    assert points_unique.is_close(points_expected)



"""Module for the Circle class."""
from __future__ import annotations

import math
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes

from skspatial._functions import np_float
from skspatial.objects._base_sphere import _BaseSphere
from skspatial.objects.line import Line
from skspatial.objects.point import Point
from skspatial.objects.points import Points
from skspatial.typing import array_like


class Circle(_BaseSphere):
    """
    A circle in 2D space.

    The circle is defined by a 2D point and a radius.

    Parameters
    ----------
    point : (2,) array_like
        Center of the circle.
    radius : {int, float}
        Radius of the circle/

    Attributes
    ----------
    point : (2,) Point
        Center of the circle.
    radius : {int, float}
        Radius of the circle.
    dimension : int
        Dimension of the circle.

    Raises
    ------
    ValueError
        If the radius is not positive.
        If the point is not 2D.

    Examples
    --------
    >>> from skspatial.objects import Circle

    >>> circle = Circle([2, 5], 3)

    >>> circle
    Circle(point=Point([2, 5]), radius=3)

    >>> circle.dimension
    2

    >>> circle.area().round(2)
    28.27

    >>> Circle([0, 0, 0], 1)
    Traceback (most recent call last):
    ...
    ValueError: The point must be 2D.

    >>> Circle([0, 0], 0)
    Traceback (most recent call last):
    ...
    ValueError: The radius must be positive.

    """

    def __init__(self, point: array_like, radius: float):

        super().__init__(point, radius)

        if self.point.dimension != 2:
            raise ValueError("The point must be 2D.")

    @np_float
    def circumference(self) -> float:
        r"""
        Return the circumference of the circle.

        The circumference :math:`C` of a circle with radius :math:`r` is

        .. math:: C = 2 \pi r

        Returns
        -------
        np.float64
            Circumference of the circle.

        Examples
        --------
        >>> from skspatial.objects import Circle

        >>> Circle([0, 0], 1).area().round(2)
        3.14

        >>> Circle([0, 0], 2).area().round(2)
        12.57

        """
        return 2 * np.pi * self.radius

    @np_float
    def area(self) -> float:
        r"""
        Return the area of the circle.

        The area :math:`A` of a circle with radius :math:`r` is

        .. math:: A = \pi r ^ 2

        Returns
        -------
        np.float64
            Area of the circle.

        Examples
        --------
        >>> from skspatial.objects import Circle

        >>> Circle([0, 0], 1).area().round(2)
        3.14

        >>> Circle([0, 0], 2).area().round(2)
        12.57

        """
        return np.pi * self.radius**2

    def intersect_line(self, line: Line) -> Tuple[Point, Point]:
        """
        Intersect the circle with a line.

        A line intersects a circle at two points.

        Parameters
        ----------
        line : Line
            Input line.

        Returns
        -------
        point_a, point_b : Point
            The two points of intersection.

        Raises
        ------
        ValueError
            If the line does not intersect the circle.

        References
        ----------
        http://mathworld.wolfram.com/Circle-LineIntersection.html

        Examples
        --------
        >>> from skspatial.objects import Circle, Line

        >>> circle = Circle([0, 0], 1)

        >>> circle.intersect_line(Line(point=[0, 0], direction=[1, 0]))
        (Point([-1.,  0.]), Point([1., 0.]))

        >>> point_a, point_b = circle.intersect_line(Line(point=[0, 0], direction=[1, 1]))

        >>> point_a.round(3)
        Point([-0.707, -0.707])

        >>> point_b.round(3)
        Point([0.707, 0.707])

        >>> circle.intersect_line(Line(point=[1, 2], direction=[1, 1]))
        (Point([-1.,  0.]), Point([0., 1.]))

        If the line is tangent to the circle, the two intersection points are the same.

        >>> circle.intersect_line(Line(point=[1, 0], direction=[0, 1]))
        (Point([1., 0.]), Point([1., 0.]))

        The circle does not have to be centered on the origin.

        >>> point_a, point_b = Circle([2, 3], 5).intersect_line(Line([1, 1], [2, 3]))

        >>> point_a.round(3)
        Point([-0.538, -1.308])

        >>> point_b.round(3)
        Point([5., 7.])

        >>> circle.intersect_line(Line(point=[5, 0], direction=[1, 1]))
        Traceback (most recent call last):
        ...
        ValueError: The line does not intersect the circle.

        """
        # Two points on the line.
        point_1 = line.point
        point_2 = point_1 + line.direction.unit()

        # Translate the points on the line to mimic the circle being centered on the origin.
        point_translated_1 = point_1 - self.point
        point_translated_2 = point_2 - self.point

        x_1, y_1 = point_translated_1
        x_2, y_2 = point_translated_2

        d_x = x_2 - x_1
        d_y = y_2 - y_1

        # Pre-compute variables common to x and y equations.
        d_r_squared = d_x**2 + d_y**2
        determinant = x_1 * y_2 - x_2 * y_1
        discriminant = self.radius**2 * d_r_squared - determinant**2

        if discriminant < 0:
            raise ValueError("The line does not intersect the circle.")

        root = math.sqrt(discriminant)

        pm = np.array([-1, 1])  # Array to compute minus/plus.
        sign = -1 if d_y < 0 else 1

        coords_x = (determinant * d_y + pm * sign * d_x * root) / d_r_squared
        coords_y = (-determinant * d_x + pm * abs(d_y) * root) / d_r_squared

        point_translated_a = Point([coords_x[0], coords_y[0]])
        point_translated_b = Point([coords_x[1], coords_y[1]])

        # Translate the intersection points back from origin circle to real circle.
        point_a = point_translated_a + self.point
        point_b = point_translated_b + self.point

        return point_a, point_b

    @classmethod
    def best_fit(cls, points: array_like) -> Circle:
        """
        Return the sphere of best fit for a set of 2D points.

        Parameters
        ----------
        points : array_like
             Input 2D points.

        Returns
        -------
        Circle
            The circle of best fit.

        Raises
        ------
        ValueError
            If the points are not 2D.
            If there are fewer than three points.
            If the points are collinear.

        Reference
        ---------
        https://meshlogic.github.io/posts/jupyter/curve-fitting/fitting-a-circle-to-cluster-of-3d-points/

        Examples
        --------
        >>> import numpy as np

        >>> from skspatial.objects import Circle

        >>> points = [[1, 1], [2, 2], [3, 1]]
        >>> circle = Circle.best_fit(points)

        >>> circle.point
        Point([2., 1.])

        >>> np.round(circle.radius, 2)
        1.0

        """
        points = Points(points)

        if points.dimension != 2:
            raise ValueError("The points must be 2D.")

        if points.shape[0] < 3:
            raise ValueError("There must be at least 3 points.")

        if points.affine_rank() != 2:
            raise ValueError("The points must not be collinear.")

        n = points.shape[0]
        A = np.hstack((2 * points, np.ones((n, 1))))
        b = (points**2).sum(axis=1)
        c = np.linalg.lstsq(A, b, rcond=None)[0]

        center = c[:2]
        radius = np.sqrt(c[2] + c[0] ** 2 + c[1] ** 2)

        return cls(center, radius)

    def plot_2d(self, ax_2d: Axes, **kwargs) -> None:
        """
        Plot the circle in 2D.

        Parameters
        ----------
        ax_2d : Axes
            Instance of :class:`~matplotlib.axes.Axes`.
        kwargs : dict, optional
            Additional keywords passed to :class:`matplotlib.patches.Circle`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt

            >>> from skspatial.objects import Circle

            >>> circle = Circle([-2, 3], 3)

            >>> fig, ax = plt.subplots()
            >>> circle.plot_2d(ax, fill=False)
            >>> circle.point.plot_2d(ax)
            >>> limits = plt.axis([-10, 10, -10, 10])

        """
        circle = plt.Circle(self.point, self.radius, **kwargs)

        ax_2d.add_artist(circle)



"""Measurements using spatial objects."""
import numpy as np

from skspatial.objects import Points
from skspatial.objects import Vector
from skspatial.typing import array_like


def area_triangle(point_a: array_like, point_b: array_like, point_c: array_like) -> np.float64:
    """
    Return the area of a triangle defined by three points.

    The points are the vertices of the triangle. They must be 3D or less.

    Parameters
    ----------
    point_a, point_b, point_c : array_like
        The three vertices of the triangle.

    Returns
    -------
    np.float64
        The area of the triangle.

    References
    ----------
    http://mathworld.wolfram.com/TriangleArea.html

    Examples
    --------
    >>> from skspatial.measurement import area_triangle

    >>> area_triangle([0, 0], [0, 1], [1, 0])
    0.5

    >>> area_triangle([0, 0], [0, 2], [1, 1])
    1.0

    >>> area_triangle([3, -5, 1], [5, 2, 1], [9, 4, 2]).round(2)
    12.54

    """
    vector_ab = Vector.from_points(point_a, point_b)
    vector_ac = Vector.from_points(point_a, point_c)

    # Normal vector of plane defined by the three points.
    vector_normal = vector_ab.cross(vector_ac)

    return 0.5 * vector_normal.norm()


def volume_tetrahedron(
    point_a: array_like,
    point_b: array_like,
    point_c: array_like,
    point_d: array_like,
) -> np.float64:
    """
    Return the volume of a tetrahedron defined by four points.

    The points are the vertices of the tetrahedron. They must be 3D or less.

    Parameters
    ----------
    point_a, point_b, point_c, point_d : array_like
        The four vertices of the tetrahedron.

    Returns
    -------
    np.float64
        The volume of the tetrahedron.

    References
    ----------
    http://mathworld.wolfram.com/Tetrahedron.html

    Examples
    --------
    >>> from skspatial.measurement import volume_tetrahedron

    >>> volume_tetrahedron([0, 0], [3, 2], [-3, 5], [1, 8])
    0.0

    >>> volume_tetrahedron([0, 0, 0], [2, 0, 0], [1, 1, 0], [0, 0, 1]).round(3)
    0.333

    >>> volume_tetrahedron([0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]).round(3)
    0.167

    """
    vector_ab = Vector.from_points(point_a, point_b)
    vector_ac = Vector.from_points(point_a, point_c)
    vector_ad = Vector.from_points(point_a, point_d)

    vector_cross = vector_ac.cross(vector_ad)

    # Set the dimension to 3 so it matches the cross product.
    vector_ab = vector_ab.set_dimension(3)

    return 1 / 6 * abs(vector_ab.dot(vector_cross))


def area_signed(points: array_like) -> float:
    """
    Return the signed area of a simple polygon given the 2D coordinates of its veritces.

    The signed area is computed using the shoelace algorithm. A positive area is
    returned for a polygon whose vertices are given by a counter-clockwise
    sequence of points.

    Parameters
    ----------
    points : array_like
         Input 2D points.

    Returns
    -------
    area_signed : float
        The signed area of the polygon.

    Raises
    ------
    ValueError
        If the points are not 2D.
        If there are fewer than three points.

    References
    ----------
    https://en.wikipedia.org/wiki/Shoelace_formula
    https://alexkritchevsky.com/2018/08/06/oriented-area.html
    https://rosettacode.org/wiki/Shoelace_formula_for_polygonal_area#Python

    Examples
    --------
    >>> from skspatial.measurement import area_signed

    >>> area_signed([[0, 0], [1, 0], [0, 1]])
    0.5

    >>> area_signed([[0, 0], [0, 1], [1, 0]])
    -0.5

    >>> area_signed([[0, 0], [0, 1], [1, 2], [2, 1], [2, 0]])
    -3.0

    """
    points = Points(points)
    n_points = points.shape[0]

    if points.dimension != 2:
        raise ValueError("The points must be 2D.")

    if n_points < 3:
        raise ValueError("There must be at least 3 points.")

    X = points[:, 0]
    Y = points[:, 1]

    indices = np.arange(n_points)
    indices_offset = indices - 1

    return 0.5 * np.sum(X[indices_offset] * Y[indices] - X[indices] * Y[indices_offset])



"""Package containing spatial objects."""
from skspatial.objects.circle import Circle
from skspatial.objects.cylinder import Cylinder
from skspatial.objects.line import Line
from skspatial.objects.plane import Plane
from skspatial.objects.point import Point
from skspatial.objects.points import Points
from skspatial.objects.sphere import Sphere
from skspatial.objects.triangle import Triangle
from skspatial.objects.vector import Vector

__all__ = ['Circle', 'Cylinder', 'Line', 'Plane', 'Point', 'Points', 'Sphere', 'Triangle', 'Vector']



import math

from hypothesis import given

from skspatial.measurement import area_triangle
from skspatial.measurement import volume_tetrahedron
from skspatial.objects import Points
from tests.property.constants import ATOL
from tests.property.strategies import arrays_fixed
from tests.property.strategies import consistent_dim


@given(consistent_dim(3 * [arrays_fixed], max_dim=3))
def test_area_triangle(arrays):

    area = area_triangle(*arrays)

    if math.isclose(area, 0):
        assert Points(arrays).are_collinear(tol=ATOL)


@given(consistent_dim(4 * [arrays_fixed], max_dim=3))
def test_volume_tetrahedron(arrays):

    volume = volume_tetrahedron(*arrays)

    if math.isclose(volume, 0):
        assert Points(arrays).are_coplanar(tol=ATOL)



from hypothesis import given

from skspatial.objects import Line
from skspatial.objects import Point
from skspatial.objects import Points
from tests.property.constants import ATOL
from tests.property.strategies import arrays_fixed
from tests.property.strategies import consistent_dim


@given(consistent_dim(3 * [arrays_fixed]))
def test_are_collinear(arrays):

    array_a, array_b, array_c = arrays

    assert Points([array_a, array_a, array_a]).are_collinear(tol=ATOL)
    assert Points([array_a, array_a, array_b]).are_collinear(tol=ATOL)

    all_different = not (
        Point(array_a).is_close(array_b, abs_tol=ATOL) or Point(array_b).is_close(array_c, abs_tol=ATOL)
    )

    if Points([array_a, array_b, array_c]).are_collinear() and all_different:

        line_ab = Line.from_points(array_a, array_b)
        line_bc = Line.from_points(array_b, array_c)

        assert line_ab.contains_point(array_c, abs_tol=ATOL)
        assert line_bc.contains_point(array_a, abs_tol=ATOL)

        assert line_ab.is_coplanar(line_bc)



"""Custom strategies for property tests."""
import hypothesis.strategies as st
from hypothesis import assume

from skspatial.objects import Circle
from skspatial.objects import Line
from skspatial.objects import Plane
from skspatial.objects import Point
from skspatial.objects import Points
from skspatial.objects import Sphere
from skspatial.objects import Triangle
from skspatial.objects import Vector
from tests.property.constants import DIM_MAX
from tests.property.constants import DIM_MIN
from tests.property.constants import FLOAT_MIN


floats = st.floats(min_value=-1e4, max_value=1e4).filter(lambda x: x == 0 or abs(x) > FLOAT_MIN)

arrays = st.lists(floats, min_size=DIM_MIN, max_size=DIM_MAX)
arrays_nonzero = arrays.filter(lambda array: any(array))

radii = floats.filter(lambda x: x > 0)


@st.composite
def arrays_fixed(draw, dim=2):
    """
    Return a strategy which generates 1D arrays with a fixed length.

    Parameters
    ----------
    dim : int, optional
        Dimension (length) of the array (default is 2).

    Returns
    -------
    LazyStrategy
        Hypothesis strategy.

    Examples
    --------
    >>> from hypothesis import find
    >>> from tests.property.strategies import arrays_fixed

    >>> find(arrays_fixed(2), lambda x: True)
    [0.0, 0.0]

    >>> find(arrays_fixed(5), lambda x: True)
    [0.0, 0.0, 0.0, 0.0, 0.0]

    """
    return draw(st.lists(floats, min_size=dim, max_size=dim))


@st.composite
def arrays_fixed_nonzero(draw, dim=2):
    """
    Return a strategy which generates nonzero 1D arrays with a fixed length.

    Parameters
    ----------
    dim : int, optional
        Dimension (length) of the array (default is 2).

    Returns
    -------
    LazyStrategy
        Hypothesis strategy.

    """
    return draw(arrays_fixed(dim).filter(lambda x: any(x)))


@st.composite
def points(draw, dim):
    """
    Return a strategy which generates Point objects.

    Parameters
    ----------
    dim : int
        Dimension of the object.

    Returns
    -------
    LazyStrategy
        Hypothesis strategy.

    Examples
    --------
    >>> from hypothesis import find
    >>> from tests.property.strategies import points

    >>> find(points(3), lambda x: True)
    Point([0., 0., 0.])

    """
    return Point(draw(arrays_fixed(dim)))


@st.composite
def vectors(draw, dim):
    """
    Return a strategy which generates Vector objects.

    Parameters
    ----------
    dim : int
        Dimension of the object.

    Returns
    -------
    LazyStrategy
        Hypothesis strategy.

    Examples
    --------
    >>> from hypothesis import find
    >>> from tests.property.strategies import vectors

    >>> find(vectors(2), lambda x: True)
    Vector([0., 0.])

    """
    return Vector(draw(arrays_fixed(dim)))


@st.composite
def vectors_nonzero(draw, dim):
    """
    Return a strategy which generates nonzero Vector objects.

    Parameters
    ----------
    dim : int
        Dimension of the object.

    Returns
    -------
    LazyStrategy
        Hypothesis strategy.

    """
    return Vector(draw(arrays_fixed_nonzero(dim)))


@st.composite
def multi_points(draw, dim):
    """
    Return a strategy which generates Points objects.

    Parameters
    ----------
    dim : int
        Dimension of the object.

    Returns
    -------
    LazyStrategy
        Hypothesis strategy.

    Examples
    --------
    >>> from hypothesis import find
    >>> from tests.property.strategies import multi_points

    >>> find(multi_points(2), lambda x: len(x) == 3)
    Points([[0., 0.],
            [0., 0.],
            [0., 0.]])

    """
    n_points = draw(st.integers(min_value=1, max_value=50))
    array_like_2d = [draw(arrays_fixed(dim)) for _ in range(n_points)]

    return Points(array_like_2d)


@st.composite
def lines_or_planes(draw, LineOrPlane, dim):
    """
    Return a strategy which generates Line or Plane objects.

    Parameters
    ----------
    LineOrPlane : class
        Line or Plane class.
    dim : int
        Dimension of the object.

    Returns
    -------
    LazyStrategy
        Hypothesis strategy.

    Examples
    --------
    >>> from hypothesis import find
    >>> from tests.property.strategies import lines_or_planes

    >>> find(lines_or_planes(Line, 2), lambda x: True)
    Line(point=Point([0., 0.]), direction=Vector([0.   , 0.001]))

    >>> find(lines_or_planes(Plane, 3), lambda x: True)
    Plane(point=Point([0., 0., 0.]), normal=Vector([0.   , 0.   , 0.001]))

    """
    array_point = draw(arrays_fixed(dim))
    array_vector = draw(arrays_fixed_nonzero(dim))

    return LineOrPlane(array_point, array_vector)


@st.composite
def lines(draw, dim):
    """
    Return a strategy which generates Line objects.

    Parameters
    ----------
    dim : int
        Dimension of the object.

    Returns
    -------
    LazyStrategy
        Hypothesis strategy.

    Examples
    --------
    >>> from hypothesis import find
    >>> from tests.property.strategies import lines

    >>> find(lines(dim=4), lambda x: x.direction.min() <= -1)
    Line(point=Point([0., 0., 0., 0.]), direction=Vector([ 0.,  0.,  0., -1.]))

    """
    return draw(lines_or_planes(Line, dim))


@st.composite
def planes(draw, dim):
    """
    Return a strategy which generates Plane objects.

    Parameters
    ----------
    dim : int
        Dimension of the object.

    Returns
    -------
    LazyStrategy
        Hypothesis strategy.

    Examples
    --------
    >>> from hypothesis import find
    >>> from tests.property.strategies import planes

    >>> find(planes(dim=3), lambda x: x.normal.norm() >= 5)
    Plane(point=Point([0., 0., 0.]), normal=Vector([0., 0., 5.]))

    """
    return draw(lines_or_planes(Plane, dim))


@st.composite
def circles(draw):
    """
    Return a strategy which generates circles.

    Returns
    -------
    LazyStrategy
        Hypothesis strategy.

    Examples
    --------
    >>> from hypothesis import find
    >>> from tests.property.strategies import circles

    >>> circle = find(circles(), lambda x: x.radius >= 1)
    >>> round(circle.radius)
    1

    """
    return Circle(draw(arrays_fixed(2)), draw(radii))


@st.composite
def spheres(draw):
    """
    Return a strategy which generates Circle objects.

    Returns
    -------
    LazyStrategy
        Hypothesis strategy.

    Examples
    --------
    >>> from hypothesis import find
    >>> from tests.property.strategies import spheres

    >>> sphere = find(spheres(), lambda x: x.radius >= 1)
    >>> round(sphere.radius)
    1

    """
    return Sphere(draw(arrays_fixed(3)), draw(radii))


@st.composite
def triangles(draw, dim):
    """
    Return a strategy which generates Triangle objects.

    Parameters
    ----------
    dim : int
        Dimension of the object.

    Returns
    -------
    LazyStrategy
        Hypothesis strategy.

    Examples
    --------
    >>> from hypothesis import find
    >>> from tests.property.strategies import triangles

    >>> find(triangles(dim=2), lambda x: True)
    Triangle(point_a=Point([0., 0.]), point_b=Point([0.   , 0.001]), point_c=Point([0.001, 0.   ]))

    """
    point_a = draw(arrays_fixed(dim))
    point_b = draw(arrays_fixed(dim))
    point_c = draw(arrays_fixed(dim))

    assume(not Points([point_a, point_b, point_c]).are_collinear(tol=1))

    return Triangle(point_a, point_b, point_c)


@st.composite
def consistent_dim(draw, strategies, min_dim=DIM_MIN, max_dim=DIM_MAX):
    """
    Return a strategy which generates multiple spatial objects with the same dimension.

    Parameters
    ----------
    strategies: sequence
        Sequence of functions that return strategies for spatial objects.
        The functions must take a dimension argument.
    min_dim, max_dim: int
        Min and max dimension of the spatial objects.

    Returns
    -------
    LazyStrategy
        Hypothesis strategy.

    Examples
    --------
    >>> from hypothesis import find
    >>> from tests.property.strategies import vectors, lines, planes, consistent_dim

    >>> find(consistent_dim([vectors, planes], min_dim=3), lambda x: True)
    [Vector([0., 0., 0.]), Plane(point=Point([0., 0., 0.]), normal=Vector([0.   , 0.   , 0.001]))]

    >>> find(consistent_dim(3 * [vectors], min_dim=3), lambda x: True)
    [Vector([0., 0., 0.]), Vector([0., 0., 0.]), Vector([0., 0., 0.])]

    """
    dim = draw(st.integers(min_value=min_dim, max_value=max_dim))

    return [draw(strategy(dim)) for strategy in strategies]



class _PlotterMixin:

    dimension: int

    def plotter(self, **kwargs):
        """Return a function that plots the object when passed a matplotlib axes."""
        if self.dimension == 2:

            if not hasattr(self, 'plot_2d'):
                raise ValueError("The object cannot be plotted in 2D.")

            return lambda ax: self.plot_2d(ax, **kwargs)

        if self.dimension == 3:

            if not hasattr(self, 'plot_3d'):
                raise ValueError("The object cannot be plotted in 3D.")

            return lambda ax: self.plot_3d(ax, **kwargs)

        raise ValueError("The dimension must be 2 or 3.")


class _BaseSpatial(_PlotterMixin):
    """Private base class for all spatial objects."""

    ...



"""Private base classes for arrays."""
import warnings
from typing import Type
from typing import TypeVar

import numpy as np

from skspatial._functions import _allclose
from skspatial.objects._base_spatial import _BaseSpatial
from skspatial.typing import array_like

# Create generic variables that can be 'Parent' or any subclass.
Array = TypeVar('Array', bound='_BaseArray')

Array1D = TypeVar('Array1D', bound='_BaseArray1D')

Array2D = TypeVar('Array2D', bound='_BaseArray2D')


class _BaseArray(np.ndarray, _BaseSpatial):
    """Private base class for spatial objects based on a single NumPy array."""

    def __new__(cls: Type[Array], array: array_like) -> Array:

        with warnings.catch_warnings():

            warnings.filterwarnings("error")

            try:
                np.array(array)

            except np.VisibleDeprecationWarning as error:
                if str(error).startswith("Creating an ndarray from ragged nested sequences"):
                    raise ValueError("The array must not contain sequences with different lengths.")

        if np.size(array) == 0:
            raise ValueError("The array must not be empty.")

        if not np.isfinite(array).all():
            raise ValueError("The values must all be finite.")

        # We cast the input array to be our class type.
        obj = np.asarray(array).view(cls)

        return obj

    def __array_wrap__(self, array: np.ndarray) -> np.ndarray:
        """
        Return regular :class:`numpy.ndarray` when default NumPy method is called.

        >>> from skspatial.objects import Vector
        >>> vector = Vector([1.234, 2.1234, 3.1234])

        >>> vector.sum().round()
        6.0

        >>> vector.mean().round(2)
        2.16

        """
        return array

    def to_array(self) -> np.ndarray:
        """
        Convert the object to a regular NumPy ndarray.

        Examples
        --------
        >>> from skspatial.objects import Point

        >>> point = Point([1, 2, 3])

        >>> point.to_array()
        array([1, 2, 3])

        """
        return np.array(self)

    def is_close(self, other: array_like, **kwargs: float) -> bool:
        """
        Check if the array is close to another.

        Parameters
        ----------
        other : array_like
            Other array.
        kwargs : dict, optional
            Additional keywords passed to :func:`math.isclose`.

        Returns
        -------
        bool
            True if the arrays are close; false otherwise.

        """
        return bool(_allclose(self, other, **kwargs).all())

    def is_equal(self, other: array_like) -> bool:
        """
        Check if the array is equal to another.

        Parameters
        ----------
        other : array_like
            Other array.

        Returns
        -------
        bool
            True if the arrays are equal; false otherwise.

        """
        return np.array_equal(self, other)

    def round(self, decimals: int = 0) -> Array:  # type: ignore[override]  # noqa: A003
        """
        Round the array to the given number of decimals.

        Refer to :func:`np.around` for the full documentation.

        Examples
        --------
        >>> from skspatial.objects import Point, Vector

        >>> Vector([1, 1, 1]).unit().round(3)
        Vector([0.577, 0.577, 0.577])

        >>> Point([1, 2, 3.532]).round(2)
        Point([1.  , 2.  , 3.53])

        """
        return np.around(self, decimals=decimals, out=self)


class _BaseArray1D(_BaseArray):
    """Private base class for spatial objects based on a single 1D NumPy array."""

    def __new__(cls: Type[Array1D], array: array_like) -> Array1D:

        obj = super().__new__(cls, array)  # pytype: disable=wrong-arg-count

        if obj.ndim != 1:
            raise ValueError("The array must be 1D.")

        obj.dimension = obj.size

        return obj

    def __array_finalize__(self, _) -> None:

        self.dimension = self.size

    def set_dimension(self: Array1D, dim: int) -> Array1D:
        """
        Set the dimension (length) of the 1D array.

        Parameters
        ----------
        dim : int
            Desired dimension.
            Must be greater than or equal to the current dimension.

        Returns
        -------
        ndarray
            (dim,) array.

        Raises
        ------
        ValueError
            If the desired dimension is less than the current dimension.

        Examples
        --------
        >>> from skspatial.objects import Point

        >>> Point([1]).set_dimension(2)
        Point([1, 0])

        >>> Point([1, 2]).set_dimension(4)
        Point([1, 2, 0, 0])

        >>> Point([1, 2, 3]).set_dimension(2)
        Traceback (most recent call last):
        ...
        ValueError: The desired dimension cannot be less than the current dimension.

        """
        if dim < self.dimension:
            raise ValueError("The desired dimension cannot be less than the current dimension.")

        n_zeros = dim - self.size
        array_padded = np.pad(self, (0, n_zeros), 'constant')

        return self.__class__(array_padded)


class _BaseArray2D(_BaseArray):
    """Private base class for spatial objects based on a single 2D NumPy array."""

    def __new__(cls: Type[Array2D], array: array_like) -> Array2D:

        obj = super().__new__(cls, array)  # pytype: disable=wrong-arg-count

        if obj.ndim != 2:
            raise ValueError("The array must be 2D.")

        obj.dimension = obj.shape[1]

        return obj

    def set_dimension(self: Array2D, dim: int) -> Array2D:
        """
        Set the dimension (width) of the 2D array.

        E.g., each row of the array represents a point in space.
        The width of the array is the dimension of the points.

        Parameters
        ----------
        dim : int
            Desired dimension.
            Must be greater than or equal to the current dimension.

        Returns
        -------
        ndarray
            (N, dim) array.

        Raises
        ------
        ValueError
            If the desired dimension is less than the current dimension.

        Examples
        --------
        >>> from skspatial.objects import Points

        >>> points = Points([[1, 0], [2, 3]])

        >>> points.set_dimension(3)
        Points([[1, 0, 0],
                [2, 3, 0]])

        >>> points.set_dimension(5)
        Points([[1, 0, 0, 0, 0],
                [2, 3, 0, 0, 0]])

        >>> Points([[1, 2, 3], [4, 5, 6]]).set_dimension(2)
        Traceback (most recent call last):
        ...
        ValueError: The desired dimension cannot be less than the current dimension.

        """
        if dim < self.dimension:
            raise ValueError("The desired dimension cannot be less than the current dimension.")

        array_padded = np.pad(self, ((0, 0), (0, dim - self.dimension)), 'constant')

        return self.__class__(array_padded)

    def __array_finalize__(self, _) -> None:

        try:
            self.dimension = self.shape[1]

        except IndexError:
            self.dimension = None



import numpy as np
import pytest

from skspatial.objects import Circle
from skspatial.objects import Cylinder
from skspatial.objects import Line
from skspatial.objects import Plane
from skspatial.objects import Point
from skspatial.objects import Points
from skspatial.objects import Sphere
from skspatial.objects import Triangle
from skspatial.objects import Vector


@pytest.mark.parametrize(
    ("obj_spatial", "repr_expected"),
    [
        (Point([0]), "Point([0])"),
        (Point([0, 0]), "Point([0, 0])"),
        (Point([0.5, 0]), "Point([0.5, 0. ])"),
        (Point([-11, 0]), "Point([-11,   0])"),
        (Vector([-11, 0]), "Vector([-11,   0])"),
        (Vector([-11.0, 0.0]), "Vector([-11.,   0.])"),
        (Vector([0, 0]), "Vector([0, 0])"),
        (Vector([0.5, 0]), "Vector([0.5, 0. ])"),
        (Points([[1.5, 2], [5, 3]]), "Points([[1.5, 2. ],\n        [5. , 3. ]])"),
        (Line([0, 0], [1, 0]), "Line(point=Point([0, 0]), direction=Vector([1, 0]))"),
        (Line([-1, 2, 3], [5, 4, 2]), "Line(point=Point([-1,  2,  3]), direction=Vector([5, 4, 2]))"),
        (Line(np.zeros(2), [1, 0]), "Line(point=Point([0., 0.]), direction=Vector([1, 0]))"),
        (Plane([0, 0], [1, 0]), "Plane(point=Point([0, 0]), normal=Vector([1, 0]))"),
        (Plane([-1, 2, 3], [5, 4, 2]), "Plane(point=Point([-1,  2,  3]), normal=Vector([5, 4, 2]))"),
        (Circle([0, 0], 1), "Circle(point=Point([0, 0]), radius=1)"),
        (Circle([0, 0], 2.5), "Circle(point=Point([0, 0]), radius=2.5)"),
        (Sphere([0, 0, 0], 1), "Sphere(point=Point([0, 0, 0]), radius=1)"),
        (
            Triangle([0, 0], [0, 1], [1, 0]),
            "Triangle(point_a=Point([0, 0]), point_b=Point([0, 1]), point_c=Point([1, 0]))",
        ),
        (Cylinder([0, 0, 0], [0, 0, 1], 1), "Cylinder(point=Point([0, 0, 0]), vector=Vector([0, 0, 1]), radius=1)"),
    ],
)
def test_repr(obj_spatial, repr_expected):

    assert repr(obj_spatial) == repr_expected



"""Private functions used for plotting spatial objects with Matplotlib."""
from typing import Callable
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from mpl_toolkits.mplot3d import Axes3D

from skspatial.typing import array_like


def _scatter_2d(ax_2d: Axes, points: array_like, **kwargs) -> None:
    """
    Plot points on a 2D scatter plot.

    Parameters
    ----------
    ax_2d : Axes
        Instance of :class:`~matplotlib.axes.Axes`.
    points : array_like
        2D points.
    kwargs : dict, optional
        Additional keywords passed to :meth:`~matplotlib.axes.Axes.scatter`.

    """
    array = np.array(points)
    ax_2d.scatter(array[:, 0], array[:, 1], **kwargs)


def _scatter_3d(ax_3d: Axes3D, points: array_like, **kwargs) -> None:
    """
    Plot points on a 3D scatter plot.

    Parameters
    ----------
    ax_3d : Axes3D
        Instance of :class:`~mpl_toolkits.mplot3d.axes3d.Axes3D`.
    points : array_like
        3D points.
    kwargs : dict, optional
        Additional keywords passed to :meth:`~mpl_toolkits.mplot3d.axes3d.Axes3D.scatter`.

    Raises
    ------
    ValueError
        If the axis is not an instance of Axes3D.

    """
    if not isinstance(ax_3d, Axes3D):
        raise ValueError("Axis must be instance of class Axes3D.")

    array = np.array(points)
    ax_3d.scatter(array[:, 0], array[:, 1], array[:, 2], **kwargs)


def _connect_points_2d(ax_2d: Axes, point_a: array_like, point_b: array_like, **kwargs) -> None:
    """
    Plot a line between two 2D points.

    Parameters
    ----------
    ax_2d : Axes
        Instance of :class:`~matplotlib.axes.Axes`.
    point_a, point_b : array_like
        The two 2D points to be connected.
    kwargs : dict, optional
        Additional keywords passed to :meth:`~matplotlib.axes.Axes.plot`.

    """
    xs = [point_a[0], point_b[0]]
    ys = [point_a[1], point_b[1]]

    ax_2d.plot(xs, ys, **kwargs)


def _connect_points_3d(ax_3d: Axes3D, point_a: array_like, point_b: array_like, **kwargs) -> None:
    """
    Plot a line between two 3D points.

    Parameters
    ----------
    ax_3d : Axes3D
        Instance of :class:`~mpl_toolkits.mplot3d.axes3d.Axes3D`.
    point_a, point_b : array_like
        The two 3D points to be connected.
    kwargs : dict, optional
        Additional keywords passed to :meth:`~mpl_toolkits.mplot3d.axes3d.Axes3D.plot`.

    Raises
    ------
    ValueError
        If the axis is not an instance of Axes3D.

    """
    if not isinstance(ax_3d, Axes3D):
        raise ValueError("Axis must be instance of class Axes3D.")

    xs = [point_a[0], point_b[0]]
    ys = [point_a[1], point_b[1]]
    zs = [point_a[2], point_b[2]]

    ax_3d.plot(xs, ys, zs, **kwargs)


def plot_2d(*plotters: Callable) -> Tuple:
    """Plot multiple spatial objects in 2D."""
    fig, ax = plt.subplots()

    for plotter in plotters:
        plotter(ax)

    return fig, ax


def plot_3d(*plotters: Callable) -> Tuple:
    """Plot multiple spatial objects in 3D."""
    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")

    for plotter in plotters:
        plotter(ax)

    return fig, ax



"""Module for private parent class of Line and Plane."""
import inspect

import numpy as np

from skspatial._functions import _contains_point
from skspatial._functions import _sum_squares
from skspatial.objects._base_spatial import _BaseSpatial
from skspatial.objects.point import Point
from skspatial.objects.vector import Vector
from skspatial.typing import array_like


class _BaseLinePlane(_BaseSpatial):
    """Private parent class for Line and Plane."""

    def __init__(self, point: array_like, vector: array_like, **kwargs):

        self.point = Point(point)
        self.vector = Vector(vector)

        if self.point.dimension != self.vector.dimension:
            raise ValueError("The point and vector must have the same dimension.")

        if self.vector.is_zero(**kwargs):
            raise ValueError("The vector must not be the zero vector.")

        self.dimension = self.point.dimension

    def __repr__(self) -> str:

        name_class = type(self).__name__
        name_vector = inspect.getfullargspec(type(self)).args[-1]

        repr_point = np.array_repr(self.point)
        repr_vector = np.array_repr(self.vector)

        return f"{name_class}(point={repr_point}, {name_vector}={repr_vector})"

    def contains_point(self, point: array_like, **kwargs: float) -> bool:
        """Check if the line/plane contains a point."""
        return _contains_point(self, point, **kwargs)

    def is_close(self, other: array_like, **kwargs: float) -> bool:
        """
        Check if the line/plane is almost equivalent to another line/plane.

        The points must be close and the vectors must be parallel.

        Parameters
        ----------
        other : object
             Line or Plane.
        kwargs : dict, optional
            Additional keywords passed to :func:`math.isclose`.

        Returns
        -------
        bool
            True if the objects are almost equivalent; false otherwise.

        Raises
        ------
        TypeError
            If the input doesn't have the same type as the object.

        Examples
        --------
        >>> from skspatial.objects import Line, Plane

        >>> line_a = Line(point=[0, 0], direction=[1, 0])
        >>> line_b = Line(point=[0, 0], direction=[-2, 0])
        >>> line_a.is_close(line_b)
        True

        >>> line_b = Line(point=[50, 0], direction=[-4, 0])
        >>> line_a.is_close(line_b)
        True

        >>> line_b = Line(point=[50, 29], direction=[-4, 0])
        >>> line_a.is_close(line_b)
        False

        >>> plane_a = Plane(point=[0, 0, 0], normal=[0, 0, 5])
        >>> plane_b = Plane(point=[23, 45, 0], normal=[0, 0, -20])
        >>> plane_a.is_close(plane_b)
        True

        >>> line_a.is_close(plane_a)
        Traceback (most recent call last):
        ...
        TypeError: The input must have the same type as the object.

        """
        if not isinstance(other, type(self)):
            raise TypeError("The input must have the same type as the object.")

        contains_point = self.contains_point(other.point, **kwargs)
        is_parallel = self.vector.is_parallel(other.vector, **kwargs)

        return contains_point and is_parallel

    def sum_squares(self, points: array_like) -> np.float64:

        return _sum_squares(self, points)



import math

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal
from numpy.testing import assert_array_equal

from skspatial.objects import Line
from skspatial.objects import Plane
from skspatial.objects import Point

INPUT_MUST_ALSO_BE_LINE = "The input must also be a line."
POINTS_MUST_NOT_BE_CONCURRENT = "The points must not be concurrent."
ARRAY_MUST_NOT_BE_EMPTY = "The array must not be empty."
LINES_MUST_HAVE_SAME_DIMENSION = "The lines must have the same dimension."
DIMENSION_MUST_NOT_BE_MORE_THAN_3D = "The line dimension cannot be greater than 3."
LINES_MUST_BE_COPLANAR = "The lines must be coplanar."
LINES_MUST_NOT_BE_PARALLEL = "The lines must not be parallel."


@pytest.mark.parametrize(
    ("array_a", "array_b", "line_expected"),
    [
        ([0, 0], [1, 0], Line([0, 0], [1, 0])),
        ([0, 0], [1, 1], Line([0, 0], [1, 1])),
        ([5, 2], [9, 2], Line([5, 2], [4, 0])),
        ([1, 1], [0, 0], Line([1, 1], [-1, -1])),
        ([0, 0], [5, 0], Line([0, 0], [5, 0])),
        ([0, 5, 9, 2], [4, 0, 0, 0], Line([0, 5, 9, 2], [4, -5, -9, -2])),
    ],
)
def test_from_points(array_a, array_b, line_expected):

    assert Line.from_points(array_a, array_b).is_close(line_expected)


@pytest.mark.parametrize(
    ("array_a", "array_b"),
    [
        ([0, 0], [0, 0]),
        ([1, 2], [1, 2]),
        ([-1, 5], [-1, 5]),
        ([5, 2, 9, 3], [5, 2, 9, 3]),
    ],
)
def test_from_points_failure(array_a, array_b):

    message_expected = "The vector must not be the zero vector."

    with pytest.raises(ValueError, match=message_expected):
        Line.from_points(array_a, array_b)


@pytest.mark.parametrize(
    ("slope", "y_intercept", "line_expected"),
    [
        (0, 0, Line([0, 0], [1, 0])),
        (0, 1, Line([0, 1], [1, 0])),
        (0, 5, Line([0, 5], [1, 0])),
        (1, 0, Line([0, 0], [1, 1])),
        (-5, 0, Line([0, 0], [1, -5])),
        # The slope has the form rise / run.
        (-2 / 7, 0, Line([0, 0], [7, -2])),
        (3 / 4, 0, Line([0, 0], [4, 3])),
        (5 / 4, 0, Line([0, 0], [4, 5])),
    ],
)
def test_from_slope(slope, y_intercept, line_expected):

    assert Line.from_slope(slope, y_intercept).is_close(line_expected)


@pytest.mark.parametrize(
    ("line", "param", "array_expected"),
    [
        (Line([0, 0], [1, 0]), 0, [0, 0]),
        (Line([0, 0], [1, 0]), 1, [1, 0]),
        (Line([0, 0], [1, 0]), 5, [5, 0]),
        (Line([0, 0], [1, 0]), -8, [-8, 0]),
        (Line([5, 2, 1], [0, 2, 0]), 0, [5, 2, 1]),
        (Line([5, 2, 1], [0, 9, 0]), 1, [5, 11, 1]),
        (Line([5, 2, 1], [0, -9, 0]), 1, [5, -7, 1]),
        (Line([6, -3, 7, 8], [0, 8, 0, 0]), 1, [6, 5, 7, 8]),
        (Line([6, -3, 7, 8], [0, 8, 0, 0]), -2, [6, -19, 7, 8]),
    ],
)
def test_to_point(line, param, array_expected):

    point = line.to_point(t=param)

    assert_array_equal(point, array_expected)


@pytest.mark.parametrize(
    ("line_a", "line_b", "bool_expected"),
    [
        (Line([0, 0], [1, 1]), Line([0, 0], [0, 1]), True),
        (Line([-6, 7], [5, 90]), Line([1, 4], [-4, 5]), True),
        (Line([0, 0, 1], [1, 1, 0]), Line([0, 0, 0], [0, 1, 0]), False),
        (Line([0, 0, 1], [1, 1, 0]), Line([0, 0, 1], [0, 1, 0]), True),
        (Line([0, 0, 1], [1, 0, 1]), Line([0, 0, 1], [2, 0, 2]), True),
    ],
)
def test_is_coplanar(line_a, line_b, bool_expected):

    assert line_a.is_coplanar(line_b) == bool_expected


@pytest.mark.parametrize(
    ("line_a", "line_b"),
    [
        (Line([0, 0], [1, 1]), Point([0, 0])),
        (Line([0, 0, 0], [1, 1, 0]), Plane([0, 0, 0], [0, 0, 1])),
    ],
)
def test_is_coplanar_failure(line_a, line_b):

    message_expected = "The input must also be a line."

    with pytest.raises(TypeError, match=message_expected):
        line_a.is_coplanar(line_b)


@pytest.mark.parametrize(
    ("point", "point_line", "vector_line", "point_expected", "dist_expected"),
    [
        ([0, 5], [0, 0], [0, 1], [0, 5], 0),
        ([0, 5], [0, 0], [0, 100], [0, 5], 0),
        ([1, 5], [0, 0], [0, 100], [0, 5], 1),
        ([0, 1], [0, 0], [1, 1], [0.5, 0.5], math.sqrt(2) / 2),
        ([1, 0], [0, 0], [1, 1], [0.5, 0.5], math.sqrt(2) / 2),
        ([0, 2], [0, 0], [1, 1], [1, 1], math.sqrt(2)),
        ([-15, 5], [0, 0], [0, 100], [0, 5], 15),
        ([50, 10], [1, -5], [0, 3], [1, 10], 49),
    ],
)
def test_project_point(point, point_line, vector_line, point_expected, dist_expected):

    line = Line(point_line, vector_line)

    point_projected = line.project_point(point)
    distance = line.distance_point(point)

    assert point_projected.is_close(point_expected)
    assert math.isclose(distance, dist_expected)


@pytest.mark.parametrize(
    ("line", "vector", "vector_expected"),
    [
        (Line([0, 0], [1, 0]), [1, 1], [1, 0]),
        (Line([-56, 72], [1, 0]), [1, 1], [1, 0]),
        (Line([-56, 72], [200, 0]), [5, 9], [5, 0]),
        (Line([-56, 72], [200, 0]), [-5, 9], [-5, 0]),
    ],
)
def test_project_vector(line, vector, vector_expected):

    vector_projected = line.project_vector(vector)
    assert vector_projected.is_close(vector_expected)


@pytest.mark.parametrize(
    ("line", "point", "value_expected"),
    [
        (Line([0, 0], [0, 1]), [0, 0], 0),
        (Line([0, 0], [0, 1]), [1, 0], 1),
        (Line([0, 0], [0, 1]), [1, 1], 1),
        (Line([0, 0], [0, 1]), [1, 10], 1),
        (Line([0, 0], [0, 1]), [1, -10], 1),
        (Line([0, 0], [0, 1]), [-1, 0], -1),
        (Line([0, 0], [0, 1]), [-1, 1], -1),
        (Line([0, 0], [0, 1]), [-1, -25], -1),
    ],
)
def test_side_point(line, point, value_expected):

    assert line.side_point(point) == value_expected


@pytest.mark.parametrize(
    ("array_point", "line", "dist_expected"),
    [
        ([0, 0], Line([0, 0], [1, 0]), 0),
        ([8, 7], Line([0, 0], [1, 0]), 7),
        ([20, -3], Line([0, 0], [1, 0]), 3),
        ([20, -3, 1], Line([0, 0, 0], [1, 0, 0]), math.sqrt(10)),
    ],
)
def test_distance_point(array_point, line, dist_expected):

    assert math.isclose(line.distance_point(array_point), dist_expected)


@pytest.mark.parametrize(
    ("line_a", "line_b", "dist_expected"),
    [
        # The lines intersect.
        (Line([10, 2], [1, 1]), Line([5, -3], [-1, 0]), 0),
        (Line([0, 0], [1, 1]), Line([1, 0], [1, 2]), 0),
        # The lines are parallel.
        (Line([0, 0], [1, 0]), Line([0, 0], [-1, 0]), 0),
        (Line([0, 0], [1, 0]), Line([0, 0], [1, 0]), 0),
        (Line([24, 0], [0, 1]), Line([3, 0], [0, -5]), 21),
        (Line([0, 0], [1, 1]), Line([1, 0], [1, 1]), math.sqrt(2) / 2),
        # The lines are skew.
        (Line([0, 0, 0], [0, 1, 0]), Line([1, 0, 0], [0, -4, 13]), 1),
    ],
)
def test_distance_line(line_a, line_b, dist_expected):

    assert math.isclose(line_a.distance_line(line_b), dist_expected)


@pytest.mark.parametrize(
    ("line_a", "line_b", "array_expected"),
    [
        (Line([0, 0], [1, 0]), Line([0, 0], [1, 1]), [0, 0]),
        (Line([0, 0], [1, 0]), Line([5, 5], [1, 1]), [0, 0]),
        (Line([0, 0], [1, 0]), Line([9, 0], [1, 1]), [9, 0]),
        (Line([0, 0], [1, 1]), Line([4, 0], [1, -1]), [2, 2]),
        (Line([0, 0, 0], [1, 1, 1]), Line([4, 4, 0], [-1, -1, 1]), [2, 2, 2]),
    ],
)
def test_intersect_line(line_a, line_b, array_expected):

    point_intersection = line_a.intersect_line(line_b)
    assert point_intersection.is_close(array_expected)


@pytest.mark.parametrize(
    ("line_a", "line_b", "message_expected"),
    [
        (
            Line([0, 0], [1, 0]),
            Line([0, 0, 0], [1, 0, 0]),
            LINES_MUST_HAVE_SAME_DIMENSION,
        ),
        (Line([0], [1]), Line([0, 0], [0, 1]), LINES_MUST_HAVE_SAME_DIMENSION),
        (
            Line([0, 0], [1, 0]),
            Line(4 * [0], [1, 0, 0, 0]),
            LINES_MUST_HAVE_SAME_DIMENSION,
        ),
        (
            Line(4 * [0], [1, 0, 0, 0]),
            Line(4 * [0], [1, 0, 0, 0]),
            DIMENSION_MUST_NOT_BE_MORE_THAN_3D,
        ),
        (Line([0, 0], [1, 0]), Line([0, 0], [1, 0]), LINES_MUST_NOT_BE_PARALLEL),
        (Line([0, 0], [1, 0]), Line([5, 5], [1, 0]), LINES_MUST_NOT_BE_PARALLEL),
        (Line([0, 0], [0, 1]), Line([0, 0], [0, 5]), LINES_MUST_NOT_BE_PARALLEL),
        (Line([0, 0], [1, 0]), Line([0, 0], [-1, 0]), LINES_MUST_NOT_BE_PARALLEL),
        (Line([0, 0], [1, 0]), Line([5, 5], [-1, 0]), LINES_MUST_NOT_BE_PARALLEL),
        (
            Line([0, 0, 0], [1, 1, 1]),
            Line([0, 1, 0], [-1, 0, 0]),
            LINES_MUST_BE_COPLANAR,
        ),
    ],
)
def test_intersect_line_failure(line_a, line_b, message_expected):

    with pytest.raises(ValueError, match=message_expected):
        line_a.intersect_line(line_b)


@pytest.mark.parametrize(
    ("line", "points", "error_expected"),
    [
        (Line([0, 0], [1, 0]), [[0, 0], [10, 0]], 0),
        (Line([0, 0], [5, 0]), [[0, 0], [0, 1]], 1),
        (Line([0, 0], [1, 0]), [[0, 1], [0, -1]], 2),
        (Line([0, 0], [1, 0]), [[0, 5]], 25),
        (Line([0, 0], [1, 0]), [[0, 3], [0, -2]], 13),
        (Line([0, 0], [-20, 0]), [[1, 3], [2, -2], [3, -5]], 38),
    ],
)
def test_sum_squares(line, points, error_expected):

    error = line.sum_squares(points)
    assert math.isclose(error, error_expected)


@pytest.mark.parametrize(
    ("points", "line_expected"),
    [
        ([[0, 0], [1, 0]], Line([0.5, 0], [1, 0])),
        ([[1, 0], [0, 0]], Line([0.5, 0], [-1, 0])),
        ([[0, 0], [10, 0]], Line([5, 0], [1, 0])),
        ([[0, 0], [-10, 0]], Line([-5, 0], [-1, 0])),
        ([[0, 0], [1, 1], [2, 2]], Line([1, 1], [1, 1])),
        ([[2, 2], [1, 1], [0, 0]], Line([1, 1], [-1, -1])),
        ([[0, 0], [0, 1], [1, 0], [1, 1]], Line([0.5, 0.5], [1, 0])),
    ],
)
def test_best_fit(points, line_expected):

    line_fit = Line.best_fit(np.array(points))

    assert line_fit.is_close(line_expected)
    assert line_fit.point.is_close(line_expected.point)


@pytest.mark.parametrize(
    ("points", "message_expected"),
    [
        ([[]], ARRAY_MUST_NOT_BE_EMPTY),
        ([[0, 0]], POINTS_MUST_NOT_BE_CONCURRENT),
        ([[0, 0, 0]], POINTS_MUST_NOT_BE_CONCURRENT),
        ([[1, 2], [1, 2], [1, 2]], POINTS_MUST_NOT_BE_CONCURRENT),
    ],
)
def test_best_fit_failure(points, message_expected):

    with pytest.raises(ValueError, match=message_expected):
        Line.best_fit(points)


@pytest.mark.parametrize(
    ("line", "points", "coords_expected"),
    [
        (Line([0, 0], [1, 0]), [[1, 0], [2, 0], [3, 0], [4, 0]], [1, 2, 3, 4]),
        # The point on the line acts as the origin.
        (Line([3, 0], [1, 0]), [[1, 0], [2, 0], [3, 0], [4, 0]], [-2, -1, 0, 1]),
        (
            Line([0, 0], [1, 1]),
            [[1, 0], [2, 0], [3, 0], [0, 1], [0, 2], [0, 3]],
            math.sqrt(2) * np.array([0.5, 1, 1.5, 0.5, 1, 1.5]),
        ),
        # The magnitude of the direction vector is irrelevant.
        (
            Line([0, 0], [3, 3]),
            [[1, 0], [2, 0], [3, 0], [0, 1], [0, 2], [0, 3]],
            math.sqrt(2) * np.array([0.5, 1, 1.5, 0.5, 1, 1.5]),
        ),
        (
            Line([0, 0, 0], [1, 0, 0]),
            [[1, 20, 3], [2, -5, 8], [3, 59, 100], [4, 0, 14]],
            [1, 2, 3, 4],
        ),
        (
            Line([0, 0, 0], [0, 1, 0]),
            [[1, 20, 3], [2, -5, 8], [3, 59, 100], [4, 0, 14]],
            [20, -5, 59, 0],
        ),
    ],
)
def test_transform_points(line, points, coords_expected):

    coordinates = line.transform_points(points)
    assert_array_almost_equal(coordinates, coords_expected)



from hypothesis import assume
from hypothesis import given

from skspatial.objects import Plane
from skspatial.objects import Points
from tests.property.constants import ATOL
from tests.property.strategies import arrays_fixed
from tests.property.strategies import consistent_dim


@given(consistent_dim(3 * [arrays_fixed], max_dim=3))
def test_from_points(arrays):

    points = Points(arrays)
    assume(not points.are_collinear(tol=1))

    # The plane must contain each point.
    plane = Plane.from_points(*points)

    points = points.set_dimension(plane.dimension)

    for point in points:
        assert plane.contains_point(point, abs_tol=ATOL)

    # The plane of best fit should be the same
    # as the plane from three points.
    plane_fit = Plane.best_fit(points)
    assert plane_fit.is_close(plane, abs_tol=ATOL)



"""Module for the Points class."""
from __future__ import annotations

from typing import cast

import numpy as np
from matplotlib.axes import Axes
from mpl_toolkits.mplot3d import Axes3D
from numpy.linalg import matrix_rank

from skspatial.objects._base_array import _BaseArray2D
from skspatial.objects.point import Point
from skspatial.plotting import _scatter_2d
from skspatial.plotting import _scatter_3d


class Points(_BaseArray2D):
    """
    Multiple points in space implemented as a 2D array.

    The array is a subclass of :class:`numpy.ndarray`.
    Each row in the array represents a point in space.

    Parameters
    ----------
    points : array_like
        (N, D) array representing N points with dimension D.

    Raises
    ------
    ValueError
        If the array is empty, the values are not finite,
        or the dimension is not two.

    Examples
    --------
    >>> from skspatial.objects import Points

    >>> points = Points([[1, 2, 0], [5, 4, 3]])

    >>> points
    Points([[1, 2, 0],
            [5, 4, 3]])

    >>> points.dimension
    3

    The object inherits methods from :class:`numpy.ndarray`.

    >>> points.mean(axis=0)
    array([3. , 3. , 1.5])

    >>> Points([])
    Traceback (most recent call last):
    ...
    ValueError: The array must not be empty.

    >>> import numpy as np

    >>> Points([[1, 2], [1, np.nan]])
    Traceback (most recent call last):
    ...
    ValueError: The values must all be finite.

    >>> Points([1, 2, 3])
    Traceback (most recent call last):
    ...
    ValueError: The array must be 2D.

    """

    def unique(self) -> Points:
        """
        Return unique points.

        The output contains the unique rows of the original array.

        Returns
        -------
        Points
            (N, D) array of N unique points with dimension D.

        Examples
        --------
        >>> from skspatial.objects import Points

        >>> points = Points([[1, 2, 3], [2, 3, 4], [1, 2, 3]])

        >>> points.unique()
        Points([[1, 2, 3],
                [2, 3, 4]])

        """
        return Points(np.unique(self, axis=0))

    def centroid(self) -> Point:
        """
        Return the centroid of the points.

        Returns
        -------
        Point
            Centroid of the points.

        Examples
        --------
        >>> from skspatial.objects import Points

        >>> Points([[1, 2, 3], [2, 2, 3]]).centroid()
        Point([1.5, 2. , 3. ])

        """
        centroid_ = cast(np.ndarray, self.mean(axis=0))

        return Point(centroid_)

    def mean_center(self, return_centroid: bool = False):
        """
        Mean-center the points by subtracting the centroid.

        Parameters
        ----------
        return_centroid : bool, optional
            If True, also return the original centroid of the points.

        Returns
        -------
        points_centered : (N, D) Points
            Array of N mean-centered points with dimension D.
        centroid : (D,) Point, optional
            Original centroid of the points. Only provided if `return_centroid` is True.

        Examples
        --------
        >>> from skspatial.objects import Points

        >>> points_centered, centroid = Points([[4, 4, 4], [2, 2, 2]]).mean_center(return_centroid=True)
        >>> points_centered
        Points([[ 1.,  1.,  1.],
                [-1., -1., -1.]])

        >>> centroid
        Point([3., 3., 3.])

        The centroid of the centered points is the origin.

        >>> points_centered.centroid()
        Point([0., 0., 0.])

        """
        centroid = self.centroid()
        points_centered = self - centroid

        if return_centroid:
            return points_centered, centroid

        return points_centered

    def normalize_distance(self) -> Points:
        """
        Normalize the distances of the points from the origin.

        The normalized points lie within a unit sphere centered on the origin.

        Returns
        -------
        Points
            Normalized points.

        Examples
        --------
        >>> from skspatial.objects import Points

        >>> points = Points([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

        >>> points.normalize_distance().round(3)
        Points([[0.072, 0.144, 0.215],
                [0.287, 0.359, 0.431],
                [0.503, 0.574, 0.646]])

        The transformation can be chained with mean centering.

        >>> points.mean_center().normalize_distance().round(3)
        Points([[-0.577, -0.577, -0.577],
                [ 0.   ,  0.   ,  0.   ],
                [ 0.577,  0.577,  0.577]])

        """
        distances_to_points = np.linalg.norm(self, axis=1)

        return self / distances_to_points.max()

    def affine_rank(self, **kwargs) -> np.int64:
        """
        Return the affine rank of the points.

        The affine rank is the dimension of the smallest affine space that contains the points.
        A rank of 1 means the points are collinear, and a rank of 2 means they are coplanar.

        Parameters
        ----------
        kwargs : dict, optional
            Additional keywords passed to :func:`numpy.linalg.matrix_rank`

        Returns
        -------
        np.int64
            Affine rank of the points.

        Examples
        --------
        >>> from skspatial.objects import Points

        >>> Points([[5, 5], [5, 5]]).affine_rank()
        0

        >>> Points([[5, 3], [-6, 20]]).affine_rank()
        1

        >>> Points([[0, 0], [1, 1], [2, 2]]).affine_rank()
        1

        >>> Points([[0, 0], [1, 0], [2, 2]]).affine_rank()
        2

        >>> Points([[0, 1, 0], [1, 1, 0], [2, 2, 2]]).affine_rank()
        2

        >>> Points([[0, 0], [0, 1], [1, 0], [1, 1]]).affine_rank()
        2

        >>> Points([[1, 3, 2], [3, 4, 5], [2, 1, 5], [5, 9, 8]]).affine_rank()
        3

        """
        # Remove duplicate points so they do not affect the centroid.
        points_centered = self.unique().mean_center()

        return matrix_rank(points_centered, **kwargs)

    def are_concurrent(self, **kwargs) -> bool:
        """
        Check if the points are all contained in one point.

        Parameters
        ----------
        kwargs : dict, optional
            Additional keywords passed to :func:`numpy.linalg.matrix_rank`

        Returns
        -------
        bool
            True if points are concurrent; false otherwise.

        Examples
        --------
        >>> from skspatial.objects import Points

        >>> Points([[0, 0], [1, 1], [1, 1]]).are_concurrent()
        False

        >>> Points([[1, 1], [1, 1], [1, 1]]).are_concurrent()
        True

        """
        return self.affine_rank(**kwargs) == 0

    def are_collinear(self, **kwargs) -> bool:
        """
        Check if the points are all contained in one line.

        Parameters
        ----------
        kwargs : dict, optional
            Additional keywords passed to :func:`numpy.linalg.matrix_rank`

        Returns
        -------
        bool
            True if points are collinear; false otherwise.

        Examples
        --------
        >>> from skspatial.objects import Points

        >>> Points(([0, 0, 0], [1, 2, 3], [2, 4, 6])).are_collinear()
        True

        >>> Points(([0, 0, 0], [1, 2, 3], [5, 2, 0])).are_collinear()
        False

        >>> Points(([0, 0], [1, 2], [5, 2], [6, 3])).are_collinear()
        False

        """
        return self.affine_rank(**kwargs) <= 1

    def are_coplanar(self, **kwargs) -> bool:
        """
        Check if the points are all contained in one plane.

        Parameters
        ----------
        kwargs : dict, optional
            Additional keywords passed to :func:`numpy.linalg.matrix_rank`

        Returns
        -------
        bool
            True if points are coplanar; false otherwise.

        Examples
        --------
        >>> from skspatial.objects import Points

        >>> Points([[1, 2], [9, -18], [12, 4], [2, 1]]).are_coplanar()
        True

        >>> Points([[1, 2], [9, -18], [12, 4], [2, 2]]).are_coplanar()
        True

        >>> Points([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]]).are_coplanar()
        False

        """
        return self.affine_rank(**kwargs) <= 2

    def plot_2d(self, ax_2d: Axes, **kwargs) -> None:
        """
        Plot the points on a 2D scatter plot.

        Parameters
        ----------
        ax_2d : Axes
            Instance of :class:`~matplotlib.axes.Axes`.
        kwargs : dict, optional
            Additional keywords passed to :meth:`~matplotlib.axes.Axes.scatter`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt

            >>> from skspatial.objects import Points

            >>> fig, ax = plt.subplots()
            >>> points = Points([[1, 2], [3, 4], [-4, 2], [-2, 3]])
            >>> points.plot_2d(ax, c='k')

        """
        _scatter_2d(ax_2d, self, **kwargs)

    def plot_3d(self, ax_3d: Axes3D, **kwargs) -> None:
        """
        Plot the points on a 3D scatter plot.

        Parameters
        ----------
        ax_3d : Axes3D
            Instance of :class:`~mpl_toolkits.mplot3d.axes3d.Axes3D`.
        kwargs : dict, optional
            Additional keywords passed to :meth:`~mpl_toolkits.mplot3d.axes3d.Axes3D.scatter`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt
            >>> from mpl_toolkits.mplot3d import Axes3D

            >>> from skspatial.objects import Points

            >>> fig = plt.figure()
            >>> ax = fig.add_subplot(111, projection='3d')

            >>> points = Points([[1, 2, 1], [3, 2, -7], [-4, 2, 2], [-2, 3, 1]])
            >>> points.plot_3d(ax, s=75, depthshade=False)

        """
        _scatter_3d(ax_3d, self, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['skspatial', 'skspatial.objects']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3,<4', 'numpy>=1.20,<2.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1,<2'],
 'base_test': ['pytest==7.1.1', 'pytest-cov==2.11.1'],
 'docs': ['Sphinx==4.2.0',
          'numpydoc==1.1.0',
          'sphinx-bootstrap-theme==0.8.0',
          'sphinx-gallery==0.9.0'],
 'pre_commit': ['pre-commit==2.11.1'],
 'property': ['hypothesis==6.8.1']}

setup_kwargs = {
    'name': 'scikit-spatial',
    'version': '6.4.1',
    'description': 'Spatial objects and computations based on NumPy arrays.',
    'long_description': '\n.. figure:: images/logo.svg\n         :align: left\n         :width: 70%\n\n.. image:: https://img.shields.io/pypi/v/scikit-spatial.svg\n         :target: https://pypi.python.org/pypi/scikit-spatial\n\n.. image:: https://anaconda.org/conda-forge/scikit-spatial/badges/version.svg\n         :target: https://anaconda.org/conda-forge/scikit-spatial\n\n.. image:: https://img.shields.io/pypi/pyversions/scikit-spatial.svg\n         :target: https://pypi.python.org/pypi/scikit-spatial\n\n.. image:: https://github.com/ajhynes7/scikit-spatial/actions/workflows/main.yml/badge.svg\n         :target: https://github.com/ajhynes7/scikit-spatial/actions/workflows/main.yml\n\n.. image:: https://results.pre-commit.ci/badge/github/ajhynes7/scikit-spatial/master.svg\n   :target: https://results.pre-commit.ci/latest/github/ajhynes7/scikit-spatial/master\n   :alt: pre-commit.ci status\n\n.. image:: https://readthedocs.org/projects/scikit-spatial/badge/?version=latest\n         :target: https://scikit-spatial.readthedocs.io/en/latest/?badge=latest\n         :alt: Documentation Status\n\n.. image:: https://codecov.io/gh/ajhynes7/scikit-spatial/branch/master/graph/badge.svg\n         :target: https://codecov.io/gh/ajhynes7/scikit-spatial\n\n|\n\nIntroduction\n------------\n\nThis package provides spatial objects based on NumPy arrays, as well as computations using these objects. The package includes computations for 2D, 3D, and higher-dimensional space.\n\nThe following spatial objects are provided:\n\n   - Point\n   - Points\n   - Vector\n   - Line\n   - Plane\n   - Circle\n   - Sphere\n   - Triangle\n   - Cylinder\n\nMost of the computations fall into the following categories:\n\n   - Measurement\n   - Comparison\n   - Projection\n   - Intersection\n   - Fitting\n   - Transformation\n\nAll spatial objects are equipped with plotting methods based on ``matplotlib``. Both 2D and 3D plotting are supported. Spatial computations can be easily visualized by plotting multiple objects at once.\n\n\nWhy this instead of ``scipy.spatial`` or ``sympy.geometry``?\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\nThis package has little to no overlap with the functionality of ``scipy.spatial``. It can be viewed as an object-oriented extension.\n\nWhile similar spatial objects and computations exist in the ``sympy.geometry`` module, ``scikit-spatial`` is based on NumPy rather than symbolic math. The primary objects of ``scikit-spatial`` (``Point``, ``Points``, and ``Vector``) are actually subclasses of the NumPy *ndarray*. This gives them all the regular functionality of the *ndarray*, plus additional methods from this package.\n\n>>> from skspatial.objects import Vector\n\n>>> vector = Vector([2, 0, 0])\n\nBehaviour inherited from NumPy:\n\n>>> vector.size\n3\n>>> vector.mean().round(3)\n0.667\n\nAdditional methods from ``scikit-spatial``:\n\n>>> vector.norm()\n2.0\n>>> vector.unit()\nVector([1., 0., 0.])\n\n``Point`` and ``Vector`` are based on a 1D NumPy array, and ``Points`` is based on a 2D NumPy array, where each row represents a point in space.  The ``Line`` and ``Plane`` objects have ``Point`` and ``Vector`` objects as attributes.\n\nNote that most methods inherited from NumPy return a regular *ndarray*, instead of the spatial object class.\n\n>>> vector.sum()\narray(2)\n\nThis is to avoid getting a spatial object with a forbidden shape, like a zero dimension ``Vector``. Trying to convert this back to a ``Vector`` causes an exception.\n\n>>> Vector(vector.sum())\nTraceback (most recent call last):\n...\nValueError: The array must be 1D.\n\n\nBecause the computations of ``scikit-spatial`` are also based on NumPy, keyword arguments can be passed to NumPy functions. For example, a tolerance can be specified while testing for collinearity. The ``tol`` keyword is passed to ``numpy.linalg.matrix_rank``.\n\n>>> from skspatial.objects import Points\n\n>>> points = Points([[1, 2, 3], [4, 5, 6], [7, 8, 8]])\n\n>>> points.are_collinear()\nFalse\n>>> points.are_collinear(tol=1)\nTrue\n\n\n\nInstallation\n------------\n\nThe package can be installed with pip.\n\n.. code-block:: bash\n\n   $ pip install scikit-spatial\n\n\nIt can also be installed with conda.\n\n.. code-block:: bash\n\n   $ conda install scikit-spatial -c conda-forge\n\n\nExample Usage\n-------------\n\nMeasurement\n~~~~~~~~~~~\n\nMeasure the cosine similarity between two vectors.\n\n>>> from skspatial.objects import Vector\n\n>>> Vector([1, 0]).cosine_similarity([1, 1]).round(3)\n0.707\n\n\nComparison\n~~~~~~~~~~\n\nCheck if multiple points are collinear.\n\n>>> from skspatial.objects import Points\n\n>>> points = Points([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])\n\n>>> points.are_collinear()\nTrue\n\n\nProjection\n~~~~~~~~~~\n\nProject a point onto a line.\n\n>>> from skspatial.objects import Line\n\n>>> line = Line(point=[0, 0, 0], direction=[1, 1, 0])\n\n>>> line.project_point([5, 6, 7])\nPoint([5.5, 5.5, 0. ])\n\n\nIntersection\n~~~~~~~~~~~~\n\nFind the intersection of two planes.\n\n>>> from skspatial.objects import Plane\n\n>>> plane_a = Plane(point=[0, 0, 0], normal=[0, 0, 1])\n>>> plane_b = Plane(point=[5, 16, -94], normal=[1, 0, 0])\n\n>>> plane_a.intersect_plane(plane_b)\nLine(point=Point([5., 0., 0.]), direction=Vector([0, 1, 0]))\n\n\nAn error is raised if the computation is undefined.\n\n>>> plane_b = Plane(point=[0, 0, 1], normal=[0, 0, 1])\n\n>>> plane_a.intersect_plane(plane_b)\nTraceback (most recent call last):\n...\nValueError: The planes must not be parallel.\n\n\nFitting\n~~~~~~~\n\nFind the plane of best fit for multiple points.\n\n>>> points = [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0]]\n\n>>> Plane.best_fit(points)\nPlane(point=Point([0.5, 0.5, 0. ]), normal=Vector([0., 0., 1.]))\n\n\nTransformation\n~~~~~~~~~~~~~~\n\nTransform multiple points to 1D coordinates along a line.\n\n>>> line = Line(point=[0, 0, 0], direction=[1, 2, 0])\n>>> points = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]\n\n>>> line.transform_points(points).round(3)\narray([ 2.236,  6.261, 10.286])\n\n\nAcknowledgment\n--------------\n\nThis package was created with Cookiecutter_ and the `audreyr/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`audreyr/cookiecutter-pypackage`: https://github.com/audreyr/cookiecutter-pypackage\n',
    'author': 'Andrew Hynes',
    'author_email': 'andrewjhynes@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ajhynes7/scikit-spatial',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Test behaviour of points with vectors."""
from hypothesis import given

from skspatial.objects import Point
from skspatial.objects import Vector
from tests.property.strategies import arrays
from tests.property.strategies import arrays_fixed
from tests.property.strategies import consistent_dim


@given(consistent_dim(2 * [arrays_fixed]))
def test_add(arrays):

    array_a, array_b = arrays

    # Add and subtract the array to obtain the same point.
    assert (Point(array_a) + array_b - array_b).is_close(array_a)


@given(arrays)
def test_is_close(array):

    vector = Vector(array)
    point = Point(array)

    assert point.size == vector.size

    assert point.is_close(vector)
    assert vector.is_close(point)

    assert point.is_close(array)
    assert vector.is_close(array)



"""Module for the Triangle class."""
import math
from itertools import combinations
from typing import Sequence

import numpy as np
from matplotlib.axes import Axes
from mpl_toolkits.mplot3d import Axes3D

from skspatial._functions import np_float
from skspatial.objects._base_spatial import _BaseSpatial
from skspatial.objects.line import Line
from skspatial.objects.point import Point
from skspatial.objects.points import Points
from skspatial.objects.vector import Vector
from skspatial.typing import array_like


class Triangle(_BaseSpatial):
    """
    A triangle represented by three points in space.

    Parameters
    ----------
    point_a, point_b, point_c : array_like
        The three vertices of the triangle.

    Attributes
    ----------
    point_a, point_b, point_c : Point
        The three vertices of the triangle.
    dimension : int
        Dimension of the triangle.

    Raises
    ------
    ValueError
        If the points do not have the same dimension.
        If the points are collinear.

    Examples
    --------
    >>> from skspatial.objects import Triangle

    >>> triangle = Triangle([0, 0], [1, 0], [0, 1])

    >>> triangle
    Triangle(point_a=Point([0, 0]), point_b=Point([1, 0]), point_c=Point([0, 1]))

    >>> triangle.dimension
    2

    >>> Triangle([0, 0, 0], [1, 0], [0, 1])
    Traceback (most recent call last):
    ...
    ValueError: The points must have the same dimension.

    >>> Triangle([0, 0], [0, 1], [0, 2])
    Traceback (most recent call last):
    ...
    ValueError: The points must not be collinear.

    """

    def __init__(self, point_a: array_like, point_b: array_like, point_c: array_like):

        self.point_a = Point(point_a)
        self.point_b = Point(point_b)
        self.point_c = Point(point_c)

        if not (self.point_a.dimension == self.point_b.dimension == self.point_c.dimension):
            raise ValueError("The points must have the same dimension.")

        if Points([self.point_a, self.point_b, self.point_c]).are_collinear():
            raise ValueError("The points must not be collinear.")

        self.dimension = self.point_a.dimension

    def __repr__(self) -> str:

        repr_a = np.array_repr(self.point_a)
        repr_b = np.array_repr(self.point_b)
        repr_c = np.array_repr(self.point_c)

        return f"Triangle(point_a={repr_a}, point_b={repr_b}, point_c={repr_c})"

    def multiple(self, name_method: str, inputs: Sequence) -> tuple:
        """
        Return multiple properties of the triangle.

        Parameters
        ----------
        name_method : str
            Name of the triangle method.
        inputs : Sequence
            Sequence of different inputs to the method.

        Returns
        -------
        tuple
            Multiple outputs from the triangle method.

        Examples
        --------
        >>> from math import degrees
        >>> from skspatial.objects import Triangle

        >>> triangle = Triangle([0, 0], [0, 1], [1, 0])

        >>> lengths = triangle.multiple('length', 'abc')
        >>> [round(x, 3) for x in lengths]
        [1.414, 1.0, 1.0]

        >>> angles = triangle.multiple('angle', 'ABC')
        >>> [round(degrees(x), 3) for x in angles]
        [90.0, 45.0, 45.0]

        """
        method = getattr(self, name_method)

        return tuple(method(x) for x in inputs)

    def normal(self) -> Vector:
        r"""
        Return a vector normal to the triangle.

        The normal vector is calculated as

        .. math::
            v_{AB} \times v_{AC}

        where :math:`v_{AB}` is the vector from vertex A to vertex B.

        Returns
        -------
        Vector
            Normal vector.

        Examples
        --------
        >>> from skspatial.objects import Triangle

        >>> Triangle([0, 0], [1, 0], [0, 1]).normal()
        Vector([0, 0, 1])

        The normal vector is not necessarily a unit vector.

        >>> Triangle([0, 0], [2, 0], [0, 2]).normal()
        Vector([0, 0, 4])

        The direction of the normal vector is dependent on the order of the vertices.

        >>> Triangle([0, 0], [0, 1], [1, 0]).normal()
        Vector([ 0,  0, -1])

        """
        vector_ab = Vector.from_points(self.point_a, self.point_b)
        vector_ac = Vector.from_points(self.point_a, self.point_c)

        return vector_ab.cross(vector_ac)

    def area(self) -> np.float64:
        """
        Return the area of the triangle.

        The points are the vertices of the triangle. They must be 3D or less.

        Returns
        -------
        np.float64
            The area of the triangle.

        References
        ----------
        http://mathworld.wolfram.com/TriangleArea.html

        Examples
        --------
        >>> from skspatial.objects import Triangle

        >>> Triangle([0, 0], [0, 1], [1, 0]).area()
        0.5

        >>> Triangle([0, 0], [0, 2], [1, 1]).area()
        1.0

        >>> Triangle([3, -5, 1], [5, 2, 1], [9, 4, 2]).area().round(2)
        12.54

        """
        return 0.5 * self.normal().norm()

    def perimeter(self) -> np.float64:
        """
        Return the perimeter of the triangle.

        Returns
        -------
        np.float64
            The perimeter of the triangle.

        Examples
        --------
        >>> from skspatial.objects import Triangle

        >>> Triangle([0, 0], [0, 1], [1, 0]).perimeter().round(3)
        3.414

        >>> Triangle([0, 1], [1, 1], [2, 0]).perimeter().round(3)
        4.65

        """
        result = np.sum(self.multiple('length', 'abc'))

        return np.float64(result)

    def point(self, vertex: str) -> Point:
        """
        Return a point (vertex) of the triangle.

        Parameters
        ----------
        vertex: str
            'A', 'B', or 'C'.

        Returns
        -------
        Point
            A vertex of the triangle.

        Raises
        ------
        ValueError
            If the vertex is not 'A', 'B', or 'C'.

        Examples
        --------
        >>> from skspatial.objects import Triangle

        >>> triangle = Triangle([0, 0], [1, 0], [0, 1])

        >>> triangle.point('A')
        Point([0, 0])

        >>> triangle.point('B')
        Point([1, 0])

        >>> triangle.point('C')
        Point([0, 1])

        >>> triangle.point('D')
        Traceback (most recent call last):
        ...
        ValueError: The vertex must be 'A', 'B', or 'C'.

        """
        if vertex == 'A':
            return self.point_a

        if vertex == 'B':
            return self.point_b

        if vertex == 'C':
            return self.point_c

        raise ValueError("The vertex must be 'A', 'B', or 'C'.")

    def line(self, side: str) -> Line:
        """
        Return the line along a side of the triangle.

        Parameters
        ----------
        side: str
            'a', 'b', or 'c'.
            Side 'a' is the side across from vertex 'A'.

        Returns
        -------
        Line
            Line along the side.

        Raises
        ------
        ValueError
            If the side is not 'a', 'b', or 'c'.

        Examples
        --------
        >>> from skspatial.objects import Triangle

        >>> triangle = Triangle([0, 0], [1, 0], [0, 1])

        >>> triangle.line('a')
        Line(point=Point([1, 0]), direction=Vector([-1,  1]))

        >>> triangle.line('b')
        Line(point=Point([0, 1]), direction=Vector([ 0, -1]))

        >>> triangle.line('c')
        Line(point=Point([0, 0]), direction=Vector([1, 0]))

        >>> triangle.line('d')
        Traceback (most recent call last):
        ...
        ValueError: The side must be 'a', 'b', or 'c'.

        """
        if side == 'a':
            point_1, point_2 = self.point_b, self.point_c

        elif side == 'b':
            point_1, point_2 = self.point_c, self.point_a

        elif side == 'c':
            point_1, point_2 = self.point_a, self.point_b

        else:
            raise ValueError("The side must be 'a', 'b', or 'c'.")

        return Line.from_points(point_1, point_2)

    def length(self, side: str) -> np.float64:
        """
        Return a side length of the triangle.

        Side 'a' is the side across from vertex 'A'.
        The side length is the distance between vertices 'B' and 'C'.

        Parameters
        ----------
        side: str
            'a', 'b', or 'c'.

        Returns
        -------
        np.float64
            Side length.

        Raises
        ------
        ValueError
            If the side is not 'a', 'b', or 'c'.

        Examples
        --------
        >>> from skspatial.objects import Triangle

        >>> triangle = Triangle([0, 0], [1, 0], [0, 1])

        >>> triangle.length('a').round(3)
        1.414
        >>> triangle.length('b')
        1.0
        >>> triangle.length('c')
        1.0

        >>> triangle.length('d')
        Traceback (most recent call last):
        ...
        ValueError: The side must be 'a', 'b', or 'c'.

        """
        return self.line(side).direction.norm()

    @np_float
    def angle(self, vertex: str) -> float:
        """
        Return an interior angle of the triangle.

        The angle is in radians.

        Parameters
        ----------
        vertex: str
            'A', 'B', or 'C'.

        Returns
        -------
        np.float64
            Interior angle.

        Raises
        ------
        ValueError
            If the vertex is not 'A', 'B', or 'C'.

        Examples
        --------
        >>> from skspatial.objects import Triangle

        >>> triangle = Triangle([0, 0], [1, 0], [0, 1])

        >>> triangle.angle('A').round(3)
        1.571
        >>> triangle.angle('B').round(3)
        0.785
        >>> triangle.angle('C').round(3)
        0.785

        >>> triangle.angle('D')
        Traceback (most recent call last):
        ...
        ValueError: The vertex must be 'A', 'B', or 'C'.

        """
        a, b, c = self.multiple('length', 'abc')

        if vertex == 'A':
            arg = (b**2 + c**2 - a**2) / (2 * b * c)

        elif vertex == 'B':
            arg = (a**2 + c**2 - b**2) / (2 * a * c)

        elif vertex == 'C':
            arg = (a**2 + b**2 - c**2) / (2 * a * b)

        else:
            raise ValueError("The vertex must be 'A', 'B', or 'C'.")

        return math.acos(arg)

    def centroid(self) -> Point:
        """
        Return the centroid of the triangle.

        Returns
        -------
        Point
            Centroid of the triangle.

        Examples
        --------
        >>> from skspatial.objects import Triangle

        >>> Triangle([0, 0], [0, 1], [1, 0]).centroid().round(3)
        Point([0.333, 0.333])

        >>> Triangle([0, 0, 0], [1, 2, 3], [4, 5, 6]).centroid().round(3)
        Point([1.667, 2.333, 3.   ])

        """
        return Points([self.point_a, self.point_b, self.point_c]).centroid()

    def altitude(self, vertex: str) -> Line:
        """
        Return the line of an altitude of the triangle.

        An altitude is a line segment through a vertex and perpendicular to the opposite side.

        Parameters
        ----------
        vertex: str
            'A', 'B', or 'C'.

        Returns
        -------
        Line
            Altitude line.

        Raises
        ------
        ValueError
            If the vertex is not 'A', 'B', or 'C'.

        Examples
        --------
        >>> from skspatial.objects import Triangle

        >>> triangle = Triangle([0, 0], [1, 1], [2, 0])

        >>> triangle.altitude('A')
        Line(point=Point([0, 0]), direction=Vector([1., 1.]))

        >>> triangle.altitude('B')
        Line(point=Point([1, 1]), direction=Vector([ 0., -1.]))

        >>> triangle.altitude('C')
        Line(point=Point([2, 0]), direction=Vector([-1.,  1.]))

        >>> triangle.altitude('D')
        Traceback (most recent call last):
        ...
        ValueError: The vertex must be 'A', 'B', or 'C'.

        """
        if vertex == 'A':
            point_proj = self.line('a').project_point(self.point_a)
            return Line.from_points(self.point_a, point_proj)

        if vertex == 'B':
            point_proj = self.line('b').project_point(self.point_b)
            return Line.from_points(self.point_b, point_proj)

        if vertex == 'C':
            point_proj = self.line('c').project_point(self.point_c)
            return Line.from_points(self.point_c, point_proj)

        raise ValueError("The vertex must be 'A', 'B', or 'C'.")

    def orthocenter(self) -> Point:
        """
        Return the orthocenter of the triangle.

        The orthocenter is the intersection point of the three altitudes.

        Returns
        -------
        Point
            Orthocenter of the triangle.

        Examples
        --------
        >>> from skspatial.objects import Triangle

        >>> Triangle([0, 0], [0, 1], [1, 0]).orthocenter()
        Point([0., 0.])

        >>> Triangle([0, 0], [1, 2], [2, 0]).orthocenter()
        Point([1. , 0.5])

        """
        line_alt_a = self.altitude('A')
        line_alt_b = self.altitude('B')

        return line_alt_a.intersect_line(line_alt_b)

    def classify(self, **kwargs: float) -> str:
        """
        Classify the triangle as equilateral, isosceles, or scalene.

        Parameters
        ----------
        kwargs : dict, optional
            Additional keywords passed to :func:`math.isclose`.

        Returns
        -------
        str
            'equilateral', 'isosceles', or 'scalene'.

        Examples
        --------
        >>> import math
        >>> from skspatial.objects import Triangle

        >>> Triangle([0, 0], [1, 0], [1, 1]).classify()
        'isosceles'

        >>> Triangle([0, 0], [1, 0], [0, 1]).classify()
        'isosceles'

        >>> Triangle([0, 0], [1, 0], [0, 2]).classify()
        'scalene'

        >>> Triangle([0, 0], [1, 0], [0.5, math.sin(math.pi / 3)]).classify()
        'equilateral'

        This triangle is approximately equilateral.

        >>> triangle = Triangle([0, 0], [1, 0], [0.5, 0.866])

        >>> triangle.classify()
        'isosceles'

        >>> triangle.classify(rel_tol=1e-3)
        'equilateral'

        """
        lengths = self.multiple('length', 'abc')

        pairs = combinations(lengths, 2)
        n_pairs_close = sum(math.isclose(a, b, **kwargs) for a, b in pairs)

        if n_pairs_close == 3:
            return 'equilateral'

        if n_pairs_close == 1:
            return 'isosceles'

        return 'scalene'

    def is_right(self, **kwargs: float) -> bool:
        """
        Check if the triangle is a right triangle.

        A right triangle with sides :math:`abc`, satisfies the Pythagorean theorem

        .. math::
            a^2 + b^2 = c^2

        Parameters
        ----------
        kwargs : dict, optional
            Additional keywords passed to :func:`math.isclose`.

        Returns
        -------
        bool
            True if the triangle is a right triangle; false otherwise.

        Examples
        --------
        >>> from skspatial.objects import Triangle

        >>> Triangle([0, 0], [0, 1], [1, 0]).is_right()
        True

        >>> Triangle([0, 0], [0, 1], [1, 1]).is_right()
        True

        >>> Triangle([0, 0], [1, 2], [2, 0]).is_right()
        False

        This triangle is approximately a right triangle.

        >>> triangle = Triangle([0, 0], [100, 0], [101, 100])

        >>> triangle.is_right()
        False

        >>> triangle.is_right(rel_tol=1e-2)
        True

        """
        a, b, c = sorted(self.multiple('length', 'abc'))

        return math.isclose(a**2 + b**2, c**2, **kwargs)

    def plot_2d(self, ax_2d: Axes, part: str = 'points', **kwargs: str) -> None:
        """
        Plot a triangle in 2D.

        Parameters
        ----------
        ax_2d : Axes
            Instance of :class:`~matplotlib.axes.Axes`.
        part : str, optional
            Part of the triangle to plot.
            Either 'points' or 'lines' (default 'points').
        kwargs : dict, optional
            Additional keywords passed to :meth:`~skspatial.objects.point.plot_2d` or
            :meth:`~skspatial.objects.line.plot_2d`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt
            >>> from skspatial.objects import Triangle

            >>> triangle = Triangle([0, 0], [1, 0], [0, 1])

            >>> _, ax = plt.subplots()

            >>> triangle.plot_2d(ax, part='points', s=100, zorder=3)
            >>> triangle.plot_2d(ax, part='lines', c='k')

        """
        if part == 'points':
            for point in self.multiple('point', 'ABC'):
                point.plot_2d(ax_2d, **kwargs)

        elif part == 'lines':
            for line in self.multiple('line', 'abc'):
                line.plot_2d(ax_2d, **kwargs)

    def plot_3d(self, ax_3d: Axes3D, part: str = 'points', **kwargs: str) -> None:
        """
        Plot a triangle in 3D.

        Parameters
        ----------
        ax_3d : Axes3D
            Instance of :class:`~mpl_toolkits.mplot3d.axes3d.Axes3D`.
        part : str, optional
            Part of the triangle to plot.
            Either 'points' or 'lines' (default 'points').
        kwargs : dict, optional
            Additional keywords passed to :meth:`~skspatial.objects.Point.plot_3d` or
            :meth:`~skspatial.objects.Line.plot_3d`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt
            >>> from skspatial.objects import Triangle

            >>> triangle = Triangle([0, 0], [1, 0], [0, 1])

            >>> _, ax = plt.subplots()

            >>> triangle.plot_2d(ax, part='points', s=100, zorder=3)
            >>> triangle.plot_2d(ax, part='lines', c='k')

        """
        if part == 'points':
            for point in self.multiple('point', 'ABC'):
                point.plot_3d(ax_3d, **kwargs)

        elif part == 'lines':
            for line in self.multiple('line', 'abc'):
                line.plot_3d(ax_3d, **kwargs)



import math

import hypothesis.strategies as st
import pytest
from hypothesis import assume
from hypothesis import given

from skspatial.objects import Vector
from tests.property.constants import ATOL
from tests.property.constants import DIM_MAX
from tests.property.constants import DIM_MIN
from tests.property.strategies import arrays_fixed
from tests.property.strategies import circles
from tests.property.strategies import lines
from tests.property.strategies import planes
from tests.property.strategies import spheres


@pytest.mark.parametrize('lines_or_planes', [lines, planes])
@given(data=st.data())
def test_project_point(lines_or_planes, data):
    """Test projecting a point onto a line or plane."""

    dim = data.draw(st.integers(min_value=DIM_MIN, max_value=DIM_MAX))

    array = data.draw(arrays_fixed(dim))
    line_or_plane = data.draw(lines_or_planes(dim))

    point_projected = line_or_plane.project_point(array)

    # The projected point should lie on the line/plane.
    assert line_or_plane.contains_point(point_projected, abs_tol=ATOL)

    # The vector from the point to its projection
    # should be perpendicular to the line/plane.
    vector_projection = Vector.from_points(array, point_projected)

    # The distance from the point to its projection
    # should equal the distance to the line/plane.
    distance_projection = vector_projection.norm()
    distance_to_object = abs(line_or_plane.distance_point(array))
    assert math.isclose(distance_to_object, distance_projection, rel_tol=0.1)

    # The distance of the projection should be the
    # shortest distance from the point to the object.
    distance_points = line_or_plane.point.distance_point(array)
    assert distance_projection < distance_points or math.isclose(distance_projection, distance_points)


@pytest.mark.parametrize('circles_or_spheres', [circles, spheres])
@given(data=st.data())
def test_project_point_circle_sphere(circles_or_spheres, data):

    circle_or_sphere = data.draw(circles_or_spheres())
    array_point = data.draw(arrays_fixed(circle_or_sphere.dimension))

    assume(not circle_or_sphere.point.is_close(array_point))

    point_projected = circle_or_sphere.project_point(array_point)

    assert math.isclose(circle_or_sphere.point.distance_point(point_projected), circle_or_sphere.radius)



"""Module for the Sphere class."""
from __future__ import annotations

import math
from typing import Tuple

import numpy as np
from mpl_toolkits.mplot3d import Axes3D

from skspatial._functions import np_float
from skspatial.objects._base_sphere import _BaseSphere
from skspatial.objects._mixins import _ToPointsMixin
from skspatial.objects.line import Line
from skspatial.objects.point import Point
from skspatial.objects.points import Points
from skspatial.objects.vector import Vector
from skspatial.typing import array_like


class Sphere(_BaseSphere, _ToPointsMixin):
    """
    A sphere in 3D space.

    The sphere is defined by a 3D point and a radius.

    Parameters
    ----------
    point : (3,) array_like
        Center of the sphere.
    radius : {int, float}
        Radius of the sphere.

    Attributes
    ----------
    point : (3,) Point
        Center of the sphere.
    radius : {int, float}
        Radius of the sphere.
    dimension : int
        Dimension of the sphere.

    Raises
    ------
    ValueError
        If the radius is not positive.
        If the point is not 3D.

    Examples
    --------
    >>> from skspatial.objects import Sphere

    >>> sphere = Sphere([1, 2, 3], 5)

    >>> sphere
    Sphere(point=Point([1, 2, 3]), radius=5)

    >>> sphere.dimension
    3

    >>> sphere.surface_area().round(2)
    314.16

    >>> Sphere([0, 0], 0)
    Traceback (most recent call last):
    ...
    ValueError: The radius must be positive.

    >>> Sphere([0, 0, 0, 0], 1)
    Traceback (most recent call last):
    ...
    ValueError: The point must be 3D.

    """

    def __init__(self, point: array_like, radius: float):

        super().__init__(point, radius)

        if self.point.dimension != 3:
            raise ValueError("The point must be 3D.")

    @np_float
    def surface_area(self) -> float:
        r"""
        Return the surface area of the sphere.

        The surface area :math:`A` of a sphere with radius :math:`r` is

        .. math:: A = 4 \pi r ^ 2

        Returns
        -------
        np.float64
            Surface area of the sphere.

        Examples
        --------
        >>> from skspatial.objects import Sphere

        >>> Sphere([0, 0, 0], 1).surface_area().round(2)
        12.57

        >>> Sphere([0, 0, 0], 2).surface_area().round(2)
        50.27

        """
        return 4 * np.pi * self.radius**2

    @np_float
    def volume(self) -> float:
        r"""
        Return the volume of the sphere.

        The volume :math:`V` of a sphere with radius :math:`r` is

        .. math:: V = \frac{4}{3} \pi r ^ 3

        Returns
        -------
        np.float64
            Volume of the sphere.

        Examples
        --------
        >>> from skspatial.objects import Sphere

        >>> Sphere([0, 0, 0], 1).volume().round(2)
        4.19

        >>> Sphere([0, 0, 0], 2).volume().round(2)
        33.51

        """
        return 4 / 3 * np.pi * self.radius**3

    def intersect_line(self, line: Line) -> Tuple[Point, Point]:
        """
        Intersect the sphere with a line.

        A line intersects a sphere at two points.

        Parameters
        ----------
        line : Line
            Input line.

        Returns
        -------
        point_a, point_b : Point
            The two points of intersection.

        Examples
        --------
        >>> from skspatial.objects import Sphere, Line

        >>> sphere = Sphere([0, 0, 0], 1)

        >>> sphere.intersect_line(Line([0, 0, 0], [1, 0, 0]))
        (Point([-1.,  0.,  0.]), Point([1., 0., 0.]))

        >>> sphere.intersect_line(Line([0, 0, 1], [1, 0, 0]))
        (Point([0., 0., 1.]), Point([0., 0., 1.]))

        >>> sphere.intersect_line(Line([0, 0, 2], [1, 0, 0]))
        Traceback (most recent call last):
        ...
        ValueError: The line does not intersect the sphere.

        """
        vector_to_line = Vector.from_points(self.point, line.point)
        vector_unit = line.direction.unit()

        dot = vector_unit.dot(vector_to_line)

        discriminant = dot**2 - (vector_to_line.norm() ** 2 - self.radius**2)

        if discriminant < 0:
            raise ValueError("The line does not intersect the sphere.")

        pm = np.array([-1, 1])  # Array to compute minus/plus.
        distances = -dot + pm * math.sqrt(discriminant)

        point_a, point_b = line.point + distances.reshape(-1, 1) * vector_unit

        return point_a, point_b

    @classmethod
    def best_fit(cls, points: array_like) -> Sphere:
        """
        Return the sphere of best fit for a set of 3D points.

        Parameters
        ----------
        points : array_like
             Input 3D points.

        Returns
        -------
        Sphere
            The sphere of best fit.

        Raises
        ------
        ValueError
            If the points are not 3D.
            If there are fewer than four points.
            If the points lie in a plane.

        Examples
        --------
        >>> import numpy as np

        >>> from skspatial.objects import Sphere

        >>> points = [[1, 0, 1], [0, 1, 1], [1, 2, 1], [1, 1, 2]]
        >>> sphere = Sphere.best_fit(points)

        >>> sphere.point
        Point([1., 1., 1.])

        >>> np.round(sphere.radius, 2)
        1.0

        """
        points = Points(points)

        if points.dimension != 3:
            raise ValueError("The points must be 3D.")

        if points.shape[0] < 4:
            raise ValueError("There must be at least 4 points.")

        if points.affine_rank() != 3:
            raise ValueError("The points must not be in a plane.")

        n = points.shape[0]
        A = np.hstack((2 * points, np.ones((n, 1))))
        b = (points**2).sum(axis=1)

        c, _, _, _ = np.linalg.lstsq(A, b, rcond=None)

        center = c[:3]
        radius = float(np.sqrt(np.dot(center, center) + c[3]))

        return cls(center, radius)

    def to_mesh(self, n_angles: int = 30) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Return coordinate matrices for the 3D surface of the sphere.

        Parameters
        ----------
        n_angles: int
            Number of angles used to generate the coordinate matrices.

        Returns
        -------
        X, Y, Z: (n_angles, n_angles) ndarray
            Coordinate matrices.

        Examples
        --------
        >>> from skspatial.objects import Sphere

        >>> X, Y, Z = Sphere([0, 0, 0], 1).to_mesh(5)

        >>> X.round(3)
        array([[ 0.   ,  0.   ,  0.   ,  0.   ,  0.   ],
               [ 0.   ,  0.707,  0.   , -0.707, -0.   ],
               [ 0.   ,  1.   ,  0.   , -1.   , -0.   ],
               [ 0.   ,  0.707,  0.   , -0.707, -0.   ],
               [ 0.   ,  0.   ,  0.   , -0.   , -0.   ]])

        >>> Y.round(3)
        array([[ 0.   ,  0.   ,  0.   ,  0.   ,  0.   ],
               [ 0.707,  0.   , -0.707, -0.   ,  0.707],
               [ 1.   ,  0.   , -1.   , -0.   ,  1.   ],
               [ 0.707,  0.   , -0.707, -0.   ,  0.707],
               [ 0.   ,  0.   , -0.   , -0.   ,  0.   ]])

        >>> Z.round(3)
        array([[ 1.   ,  1.   ,  1.   ,  1.   ,  1.   ],
               [ 0.707,  0.707,  0.707,  0.707,  0.707],
               [ 0.   ,  0.   ,  0.   ,  0.   ,  0.   ],
               [-0.707, -0.707, -0.707, -0.707, -0.707],
               [-1.   , -1.   , -1.   , -1.   , -1.   ]])

        """
        angles_a = np.linspace(0, np.pi, n_angles)
        angles_b = np.linspace(0, 2 * np.pi, n_angles)

        sin_angles_a = np.sin(angles_a)
        cos_angles_a = np.cos(angles_a)

        sin_angles_b = np.sin(angles_b)
        cos_angles_b = np.cos(angles_b)

        X = self.point[0] + self.radius * np.outer(sin_angles_a, sin_angles_b)
        Y = self.point[1] + self.radius * np.outer(sin_angles_a, cos_angles_b)
        Z = self.point[2] + self.radius * np.outer(cos_angles_a, np.ones_like(angles_b))

        return X, Y, Z

    def plot_3d(self, ax_3d: Axes3D, n_angles: int = 30, **kwargs) -> None:
        """
        Plot the sphere in 3D.

        Parameters
        ----------
        ax_3d : Axes3D
            Instance of :class:`~mpl_toolkits.mplot3d.axes3d.Axes3D`.
        kwargs : dict, optional
            Additional keywords passed to :meth:`~mpl_toolkits.mplot3d.axes3d.Axes3D.plot_surface`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt
            >>> from mpl_toolkits.mplot3d import Axes3D

            >>> from skspatial.objects import Sphere

            >>> fig = plt.figure()
            >>> ax = fig.add_subplot(111, projection='3d')

            >>> sphere = Sphere([1, 2, 3], 2)

            >>> sphere.plot_3d(ax, alpha=0.2)
            >>> sphere.point.plot_3d(ax, s=100)

        """
        X, Y, Z = self.to_mesh(n_angles)

        ax_3d.plot_surface(X, Y, Z, **kwargs)



import pytest

from skspatial.objects import Line
from skspatial.objects import Plane


@pytest.mark.parametrize("class_spatial", [Line, Plane])
@pytest.mark.parametrize(
    ("point", "vector", "dim_expected"),
    [([0, 0], [1, 0], 2), ([0, 0, 0], [1, 0, 0], 3), ([0, 0, 0, 0], [1, 0, 0, 0], 4)],
)
def test_dimension(class_spatial, point, vector, dim_expected):

    object_spatial = class_spatial(point, vector)
    assert object_spatial.dimension == dim_expected
